"""
Created on Feb 4, 2016

@author: Edward Barnard
"""

import time
from typing import Tuple

import numpy as np
import pyqtgraph as pg
from qtpy import QtCore

from ScopeFoundry import LQRange, Measurement
from ScopeFoundry.helper_funcs import load_qt_ui_file, sibling_path
from ScopeFoundry.logged_quantity.collection import LQCollection

from .raster_3d_scan_base import Ui_SimpleXYScanWindow as UI_Simple3DScanWindow
from .raster_scan_base import Ui_SimpleXYScanWindow


class BaseRaster2DScan(Measurement):
    name = "base_raster_2D_scan"

    def __init__(
        self,
        app,
        h_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        v_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        h_unit: str = "",
        v_unit: str = "",
        h_spinbox_decimals: int = 4,
        v_spinbox_decimals: int = 4,
        h_spinbox_step: float = 0.1,
        v_spinbox_step: float = 0.1,
        use_external_range_sync: bool = False,
        circ_roi_size: float = 1.0,
    ):
        self.h_spinbox_decimals = h_spinbox_decimals
        self.v_spinbox_decimals = v_spinbox_decimals
        self.h_spinbox_step = h_spinbox_step
        self.v_spinbox_step = v_spinbox_step
        self.h_limits = h_limits
        self.v_limits = v_limits
        self.h_unit = h_unit
        self.v_unit = v_unit
        self.use_external_range_sync = use_external_range_sync
        self.circ_roi_size = circ_roi_size
        self.img_items = []
        self.current_scan_index = (0, 0, 0)
        Measurement.__init__(self, app)

    def setup(self):
        self.base_ui_filename = sibling_path(__file__, "raster_scan_base.ui")
        self.ui: Ui_SimpleXYScanWindow = load_qt_ui_file(self.base_ui_filename)
        self.ui.h_axis_comboBox.setVisible(False)
        self.ui.v_axis_comboBox.setVisible(False)
        self.ui.setWindowTitle(self.name)

        self.display_update_period = 0.010  # seconds

        # connect events

        # local logged quantities
        h_lq_params = dict(
            vmin=self.h_limits[0],
            vmax=self.h_limits[1],
            unit=self.h_unit,
            spinbox_decimals=self.h_spinbox_decimals,
            spinbox_step=self.h_spinbox_step,
            dtype=float,
            ro=False,
        )
        h_range = self.h_limits[1] - self.h_limits[0]
        self.h0 = self.settings.New(
            "h0", initial=self.h_limits[0] + h_range * 0.25, **h_lq_params
        )
        self.h1 = self.settings.New(
            "h1", initial=self.h_limits[0] + h_range * 0.75, **h_lq_params
        )
        v_lq_params = dict(
            vmin=self.v_limits[0],
            vmax=self.v_limits[1],
            unit=self.v_unit,
            spinbox_decimals=self.v_spinbox_decimals,
            spinbox_step=self.v_spinbox_step,
            dtype=float,
            ro=False,
        )
        v_range = self.v_limits[1] - self.v_limits[0]
        self.v0 = self.settings.New(
            "v0", initial=self.v_limits[0] + v_range * 0.25, **v_lq_params
        )
        self.v1 = self.settings.New(
            "v1", initial=self.v_limits[0] + v_range * 0.75, **v_lq_params
        )

        lq_params = dict(
            dtype=float, vmin=1e-9, vmax=abs(h_range), ro=False, unit=self.h_unit
        )
        self.dh = self.settings.New("dh", initial=self.h_spinbox_step, **lq_params)
        self.dh.spinbox_decimals = self.h_spinbox_decimals
        lq_params = dict(
            dtype=float, vmin=1e-9, vmax=abs(v_range), ro=False, unit=self.v_unit
        )
        self.dv = self.settings.New("dv", initial=self.v_spinbox_step, **lq_params)
        self.dv.spinbox_decimals = self.v_spinbox_decimals

        self.Nh = self.settings.New("Nh", initial=11, vmin=1, dtype=int, ro=False)
        self.Nv = self.settings.New("Nv", initial=11, vmin=1, dtype=int, ro=False)

        self.h_center = self.settings.New(
            "h_center",
            dtype=float,
            ro=False,
            unit=self.h_unit,
            spinbox_decimals=self.h_spinbox_decimals,
            spinbox_step=self.h_spinbox_step,
        )
        self.v_center = self.settings.New(
            "v_center",
            dtype=float,
            ro=False,
            unit=self.v_unit,
            spinbox_decimals=self.v_spinbox_decimals,
            spinbox_step=self.v_spinbox_step,
        )

        self.h_span = self.settings.New(
            "h_span",
            dtype=float,
            ro=False,
            spinbox_decimals=self.h_spinbox_decimals,
            spinbox_step=self.h_spinbox_step,
            unit=self.h_unit,
        )
        self.v_span = self.settings.New(
            "v_span",
            dtype=float,
            ro=False,
            spinbox_decimals=self.v_spinbox_decimals,
            spinbox_step=self.v_spinbox_step,
            unit=self.v_unit,
        )

        self.Npixels = self.Nh.val * self.Nv.val

        self.scan_type = self.settings.New(
            "scan_type",
            dtype=str,
            initial="raster",
            choices=(
                "raster",
                "serpentine",
                "trace_retrace",
                "ortho_raster",
                "ortho_trace_retrace",
            ),
        )

        self.continuous_scan = self.settings.New(
            "continuous_scan", dtype=bool, initial=False
        )
        self.settings.New("save_h5", dtype=bool, initial=True, ro=False)

        self.settings.New("show_previous_scans", dtype=bool, initial=True)

        self.settings.New("n_frames", dtype=int, initial=1, vmin=1)

        self.settings.New(
            "pixel_time", dtype=float, ro=True, si=True, initial=1, unit="s"
        )
        self.settings.New("line_time", dtype=float, ro=True, si=True, unit="s")
        self.settings.New("frame_time", dtype=float, ro=True, si=True, unit="s")
        self.settings.New("total_time", dtype=float, ro=True, si=True, unit="s")

        for lq_name in ["Nh", "Nv", "pixel_time", "n_frames"]:
            self.settings.get_lq(lq_name).add_listener(self.compute_times)

        self.compute_times()

        if not self.use_external_range_sync:
            self.h_range = LQRange(
                self.h0, self.h1, self.dh, self.Nh, self.h_center, self.h_span
            )
            self.v_range = LQRange(
                self.v0, self.v1, self.dv, self.Nv, self.v_center, self.v_span
            )

        for s in "h0 h1 dh v0 v1 dv".split():
            self.settings.get_lq(s).add_listener(self.compute_scan_params)

        self.scan_type.updated_value.connect(self.compute_scan_params)

        # connect events
        self.settings.get_lq("show_previous_scans").add_listener(
            self.show_hide_previous_scans,
            argtype=(bool),
        )

        self.activation.connect_to_pushButton(self.ui.start_pushButton)

        self.h0.connect_to_widget(self.ui.h0_doubleSpinBox)
        self.h1.connect_to_widget(self.ui.h1_doubleSpinBox)
        self.v0.connect_to_widget(self.ui.v0_doubleSpinBox)
        self.v1.connect_to_widget(self.ui.v1_doubleSpinBox)
        self.dh.connect_to_widget(self.ui.dh_doubleSpinBox)
        self.dv.connect_to_widget(self.ui.dv_doubleSpinBox)
        self.Nh.connect_to_widget(self.ui.Nh_doubleSpinBox)
        self.Nv.connect_to_widget(self.ui.Nv_doubleSpinBox)
        self.scan_type.connect_to_widget(self.ui.scan_type_comboBox)

        self.progress.connect_to_widget(self.ui.progressBar)

        self.settings.get_lq("continuous_scan").connect_to_widget(
            self.ui.continuous_scan_checkBox
        )
        self.settings.get_lq("save_h5").connect_to_widget(self.ui.save_h5_checkBox)

        self.settings.get_lq("show_previous_scans").connect_to_widget(
            self.ui.show_previous_scans_checkBox
        )

        self.initial_scan_setup_plotting = False
        self.scan_specific_setup()

        self.add_operation("clear_previous_scans", self.clear_previous_scans)

        self.ui.clear_previous_scans_pushButton.clicked.connect(
            self.clear_previous_scans
        )

        self.compute_scan_params()

    def set_details_widget(self, widget=None, ui_filename=None):
        # print('LOADING DETAIL UI')
        if ui_filename is not None:
            details_ui = load_qt_ui_file(ui_filename)
        if widget is not None:
            details_ui = widget
        if hasattr(self, "details_ui"):
            if self.details_ui is not None:
                self.details_ui.deleteLater()
                self.ui.details_groupBox.layout().removeWidget(self.details_ui)
                # self.details_ui.hide()
                del self.details_ui
        self.details_ui = details_ui
        # return replace_widget_in_layout(self.ui.details_groupBox,details_ui)
        self.ui.details_groupBox.layout().addWidget(self.details_ui)
        return self.details_ui

    def set_h_limits(self, vmin, vmax, set_scan_to_max=False):
        self.settings.get_lq("h0").change_min_max(vmin, vmax)
        self.settings.get_lq("h1").change_min_max(vmin, vmax)
        if set_scan_to_max:
            self.settings["h0"] = vmin
            self.settings["h1"] = vmax

    def set_v_limits(self, vmin, vmax, set_scan_to_max=False):
        self.settings.get_lq("v0").change_min_max(vmin, vmax)
        self.settings.get_lq("v1").change_min_max(vmin, vmax)
        if set_scan_to_max:
            self.settings["v0"] = vmin
            self.settings["v1"] = vmax

    def compute_scan_params(self):
        self.log.debug("compute_scan_params")
        # Don't recompute if a scan is running!
        if self.is_measuring():
            return  # maybe raise error

        # self.h_array = self.h_range.array #np.arange(self.h0.val, self.h1.val, self.dh.val, dtype=float)
        # self.v_array = self.v_range.array #np.arange(self.v0.val, self.v1.val, self.dv.val, dtype=float)

        # self.Nh.update_value(len(self.h_array))
        # self.Nv.update_value(len(self.v_array))

        self.range_extent = [self.h0.val, self.h1.val, self.v0.val, self.v1.val]

        # self.corners =  [self.h_array[0], self.h_array[-1], self.v_array[0], self.v_array[-1]]
        self.corners = self.range_extent

        self.imshow_extent = [
            self.h0.val - 0.5 * self.dh.val,
            self.h1.val + 0.5 * self.dh.val,
            self.v0.val - 0.5 * self.dv.val,
            self.v1.val + 0.5 * self.dv.val,
        ]

        self.compute_times()

        # call appropriate scan generator to determine scan size, don't compute scan arrays yet
        getattr(self, "gen_%s_scan" % self.scan_type.val)(gen_arrays=False)

    def compute_scan_arrays(self):
        self.compute_scan_params()
        gen_func_name = "gen_%s_scan" % self.scan_type.val
        print("gen_arrays:", gen_func_name)
        # calls correct scan generator function
        getattr(self, gen_func_name)(gen_arrays=True)

    def create_empty_scan_arrays(self):
        self.scan_h_positions = np.zeros(self.Npixels, dtype=float)
        self.scan_v_positions = np.zeros(self.Npixels, dtype=float)
        self.scan_slow_move = np.zeros(self.Npixels, dtype=bool)
        self.scan_index_array = np.zeros((self.Npixels, 3), dtype=int)

    def pre_run(self):
        # set all logged quantities read only
        for lqname in "h0 h1 v0 v1 dh dv Nh Nv".split():
            self.settings.get_lq(lqname).change_readonly(True)

        self.compute_scan_params()

    def post_run(self):
        # set all logged quantities writable
        for lqname in "h0 h1 v0 v1 dh dv Nh Nv".split():
            self.settings.get_lq(lqname).change_readonly(False)

    def clear_qt_attr(self, attr_name):
        if hasattr(self, attr_name):
            attr = getattr(self, attr_name)
            attr.deleteLater()
            del attr

    def setup_figure(self):
        self.compute_scan_params()

        self.clear_qt_attr("graph_layout")
        self.graph_layout = pg.GraphicsLayoutWidget(border=(100, 100, 100))
        self.ui.plot_groupBox.layout().addWidget(self.graph_layout)

        self.clear_qt_attr("img_plot")
        self.img_plot: pg.PlotItem = self.graph_layout.addPlot()

        self.img_items = []

        self.img_item = pg.ImageItem()
        self.img_items.append(self.img_item)

        self.img_plot.addItem(self.img_item)
        self.img_plot.showGrid(x=True, y=True)
        self.img_plot.setAspectLocked(lock=True, ratio=1)

        self.hist_lut = pg.HistogramLUTItem()
        self.graph_layout.addItem(self.hist_lut)

        # self.clear_qt_attr('current_stage_pos_arrow')
        self.current_stage_pos_arrow = pg.ArrowItem()
        self.current_stage_pos_arrow.setZValue(100)
        self.img_plot.addItem(self.current_stage_pos_arrow)

        self.graph_layout.nextRow()
        self.pos_label = pg.LabelItem(justify="right")
        self.pos_label.setText("=====")
        self.graph_layout.addItem(self.pos_label)

        self.scan_roi = pg.ROI([0, 0], [1, 1], movable=True)
        self.scan_roi.addScaleHandle([1, 1], [0, 0])
        self.scan_roi.addScaleHandle([0, 0], [1, 1])
        self.update_scan_roi()
        self.scan_roi.sigRegionChangeFinished.connect(self.mouse_update_scan_roi)

        self.img_plot.addItem(self.scan_roi)
        for lqname in "h0 h1 v0 v1 dh dv".split():
            self.settings.get_lq(lqname).updated_value.connect(self.update_scan_roi)

        self.img_plot.scene().sigMouseMoved.connect(self.mouseMoved)

        # GoTo position context menu
        # self.goto_cmenu_action = QtWidgets.QAction("GoTo Position", self.img_plot.scene())
        # self.img_plot.scene().contextMenu.append(self.goto_cmenu_action)
        # self.goto_cmenu_action.triggered.connect(self.on_goto_position)

        # Point ROI
        self.pt_roi = pg.CircleROI(
            (0, 0), (self.circ_roi_size, self.circ_roi_size), movable=True, pen=(0, 9)
        )
        # self.pt_roi.removeHandle(self.pt_roi.getHandles()[0])

        h = self.pt_roi.addTranslateHandle((0.5, 0.5))

        h.pen = pg.mkPen("r")
        h.update()
        self.img_plot.addItem(self.pt_roi)
        self.pt_roi.removeHandle(0)
        # self.pt_roi_plotline = pg.PlotCurveItem([0], pen=(0,9))
        # self.imview.getView().addItem(self.pt_roi_plotline)
        self.pt_roi.sigRegionChangeFinished[object].connect(self.on_update_pt_roi)

        self.connect_pos_widgets()

    def on_update_pt_roi(self, roi=None):
        if roi is None:
            roi = self.circ_roi
        roi_state = roi.saveState()
        x0, y0 = roi_state["pos"]
        xc = x0 + self.circ_roi_size / 2.0
        yc = y0 + self.circ_roi_size / 2.0
        self.new_pt_pos(xc, yc)

    def new_pt_pos(self, x, y):
        print("new_pt_pos", x, y)

    def mouse_update_scan_roi(self):
        x0, y0 = self.scan_roi.pos()
        w, h = self.scan_roi.size()
        self.h_center.update_value(x0 + w / 2)
        self.v_center.update_value(y0 + h / 2)
        self.h_span.update_value(w - self.dh.val)
        self.v_span.update_value(h - self.dv.val)
        self.compute_scan_params()
        self.update_scan_roi()

    def update_scan_roi(self):
        self.log.debug("update_scan_roi")
        x0, x1, y0, y1 = self.imshow_extent
        self.scan_roi.blockSignals(True)
        self.scan_roi.setPos((x0, y0, 0))
        self.scan_roi.setSize((x1 - x0, y1 - y0, 0))
        self.scan_roi.blockSignals(False)

    def disconnect_pos_widgets(self):
        if hasattr(self, "stage"):
            s: LQCollection = self.stage.settings
            s.get_lq("x_position").disconnect_from_widget(self.ui.x_doubleSpinBox)
            s.get_lq("y_position").disconnect_from_widget(self.ui.y_doubleSpinBox)

    def connect_pos_widgets(self):
        self.disconnect_pos_widgets()
        if hasattr(self, "stage"):
            s: LQCollection = self.stage.settings

            s.get_lq("x_position").updated_value.connect(
                self.update_arrow_pos, QtCore.Qt.UniqueConnection
            )
            s.get_lq("y_position").updated_value.connect(
                self.update_arrow_pos, QtCore.Qt.UniqueConnection
            )

            s.get_lq("x_position").connect_to_widget(self.ui.x_doubleSpinBox)
            s.get_lq("y_position").connect_to_widget(self.ui.y_doubleSpinBox)

    def update_arrow_pos(self):
        if hasattr(self, "stage"):
            x = self.stage.settings["x_position"]
            y = self.stage.settings["y_position"]
            self.current_stage_pos_arrow.setPos(x, y)

    def on_goto_position(self):
        pass

    def update_display(self):
        # self.log.debug('update_display')
        if self.initial_scan_setup_plotting:
            self.img_item = pg.ImageItem()
            self.img_items.append(self.img_item)
            self.img_plot.addItem(self.img_item)
            self.hist_lut.setImageItem(self.img_item)

            self.img_item.setImage(self.display_image_map[0, :, :])
            x0, x1, y0, y1 = self.imshow_extent
            self.log.debug(f"update_display set bounds {x0} {x1} {y0} {y1}")
            self.img_item_rect = QtCore.QRectF(x0, y0, x1 - x0, y1 - y0)
            self.img_item.setRect(self.img_item_rect)
            self.log.debug(f"update_display set bounds {self.img_item_rect}")

            self.initial_scan_setup_plotting = False
        else:
            # if self.settings.scan_type.val in ['raster']
            kk, jj, ii = self.current_scan_index
            self.disp_img = self.display_image_map[kk, :, :].T
            self.img_item.setImage(self.disp_img, autoRange=False, autoLevels=True)
            # Important to set rectangle after setImage for non-square pixels
            self.img_item.setRect(self.img_item_rect)
            self.update_LUT()

    def update_LUT(self):
        """override this function to control display LUT scaling"""
        self.hist_lut.imageChanged(autoLevel=False)

    def show_hide_previous_scans(self, show):
        print("show_hide_previous_scans", show)
        if len(self.img_items) < 2:
            return
        for img_item in self.img_items[:-1]:
            img_item.setVisible(show)

    def clear_previous_scans(self):
        # current_img = img_items.pop()
        for img_item in self.img_items[:-1]:
            print("removing", img_item)
            self.img_plot.removeItem(img_item)
            img_item.deleteLater()

        self.img_items = [self.img_item]

    def mouseMoved(self, evt):
        mousePoint = self.img_plot.vb.mapSceneToView(evt)
        # print mousePoint

        # self.pos_label_text = "H {:+02.2f} um [{}], V {:+02.2f} um [{}]: {:1.2e} Hz ".format(
        #                mousePoint.x(), ii, mousePoint.y(), jj,
        #                self.count_rate_map[jj,ii]
        #                )

        self.pos_label.setText(
            f"H {mousePoint.x():+02.2f} um [{0}], V {mousePoint.y():+02.2f} um [{0}]: {0:1.2e} Hz"
        )

    def scan_specific_setup(self):
        "subclass this function to setup additional logged quantities and gui connections"
        pass
        # self.stage = self.app.hardware.dummy_xy_stage
        # self.stage.settings.get_lq("x_position").connect_to_widget(self.ui.x_doubleSpinBox)
        # self.stage.settings.get_lq("y_position").connect_to_widget(self.ui.y_doubleSpinBox)

        # logged quantities
        # connect events

    def initialize_controller(self):
        #        self.controller = self.app.hardware['xbox_controller']
        #
        #         if hasattr(self, 'controller'):
        #             self.pt_roi.sigRegionChangeFinished.connect(self.on_update_pt_roi)
        pass

    def update_point_roi_xbox(self):
        """Not yet implemented."""
        dx = self.controller.settings["Axis_4"]
        dy = self.controller.settings["Axis_3"]
        x, y = self.pt_roi.pos()
        if abs(dx) < 0.25:
            dx = 0
        if abs(dy) < 0.25:
            dy = 0
        if dx != 0 or dy != 0:
            c = self.controller.settings["sensitivity"]
            self.pt_roi.setPos(x + (c * dx), y + (c * dy))

    @property
    def h_array(self):
        # return self.h_range.array
        return np.linspace(self.h0.val, self.h1.val, self.Nh.val)

    @property
    def v_array(self):
        # return self.v_range.array
        return np.linspace(self.v0.val, self.v1.val, self.Nv.val)

    def compute_times(self):
        # self.settings['pixel_time'] = 1.0/self.scanDAQ.settings['dac_rate']
        S = self.settings
        S["line_time"] = S["pixel_time"] * S["Nh"]
        S["frame_time"] = S["pixel_time"] * self.Npixels
        S["total_time"] = S["frame_time"] * S["n_frames"]

    #### Scan Generators
    def gen_raster_scan(self, gen_arrays=True):
        self.Npixels = self.Nh.val * self.Nv.val
        self.scan_shape = (1, self.Nv.val, self.Nh.val)

        if gen_arrays:

            self.create_empty_scan_arrays()

            t0 = time.time()

            H, V = np.meshgrid(self.h_array, self.v_array)
            self.scan_h_positions[:] = H.flat
            self.scan_v_positions[:] = V.flat

            self.scan_slow_move[:: self.Nh.val] = True

            II, JJ = np.meshgrid(np.arange(self.Nh.val), np.arange(self.Nv.val))
            self.scan_index_array[:, 1] = JJ.flat
            self.scan_index_array[:, 2] = II.flat
            # self.scan_v_positions
            print("array flatten raster gen", time.time() - t0)

    def gen_serpentine_scan(self, gen_arrays=True):
        self.Npixels = self.Nh.val * self.Nv.val
        self.scan_shape = (1, self.Nv.val, self.Nh.val)

        if gen_arrays:
            self.create_empty_scan_arrays()
            pixel_i = 0
            for jj in range(self.Nv.val):
                self.scan_slow_move[pixel_i] = True

                if jj % 2:  # odd lines
                    h_line_indicies = range(self.Nh.val)[::-1]
                else:  # even lines -- traverse in opposite direction
                    h_line_indicies = range(self.Nh.val)

                for ii in h_line_indicies:
                    self.scan_v_positions[pixel_i] = self.v_array[jj]
                    self.scan_h_positions[pixel_i] = self.h_array[ii]
                    self.scan_index_array[pixel_i, :] = [0, jj, ii]
                    pixel_i += 1

    def gen_trace_retrace_scan(self, gen_arrays=True):
        self.Npixels = 2 * self.Nh.val * self.Nv.val
        self.scan_shape = (2, self.Nv.val, self.Nh.val)

        if gen_arrays:
            self.create_empty_scan_arrays()
            pixel_i = 0
            for jj in range(self.Nv.val):
                self.scan_slow_move[pixel_i] = True
                for kk, step in [(0, 1), (1, -1)]:  # trace kk =0, retrace kk=1
                    h_line_indicies = range(self.Nh.val)[::step]
                    for ii in h_line_indicies:
                        self.scan_v_positions[pixel_i] = self.v_array[jj]
                        self.scan_h_positions[pixel_i] = self.h_array[ii]
                        self.scan_index_array[pixel_i, :] = [kk, jj, ii]
                        pixel_i += 1

    def gen_ortho_raster_scan(self, gen_arrays=True):
        self.Npixels = 2 * self.Nh.val * self.Nv.val
        self.scan_shape = (2, self.Nv.val, self.Nh.val)

        if gen_arrays:
            self.create_empty_scan_arrays()
            pixel_i = 0
            for jj in range(self.Nv.val):
                self.scan_slow_move[pixel_i] = True
                for ii in range(self.Nh.val):
                    self.scan_v_positions[pixel_i] = self.v_array[jj]
                    self.scan_h_positions[pixel_i] = self.h_array[ii]
                    self.scan_index_array[pixel_i, :] = [0, jj, ii]
                    pixel_i += 1
            for ii in range(self.Nh.val):
                self.scan_slow_move[pixel_i] = True
                for jj in range(self.Nv.val):
                    self.scan_v_positions[pixel_i] = self.v_array[jj]
                    self.scan_h_positions[pixel_i] = self.h_array[ii]
                    self.scan_index_array[pixel_i, :] = [1, jj, ii]
                    pixel_i += 1

    def gen_ortho_trace_retrace_scan(self, gen_arrays=True):
        print("gen_ortho_trace_retrace_scan")
        self.Npixels = 4 * len(self.h_array) * len(self.v_array)
        self.scan_shape = (4, self.Nv.val, self.Nh.val)

        if gen_arrays:
            self.create_empty_scan_arrays()
            pixel_i = 0
            for jj in range(self.Nv.val):
                self.scan_slow_move[pixel_i] = True
                for kk, step in [(0, 1), (1, -1)]:  # trace kk =0, retrace kk=1
                    h_line_indicies = range(self.Nh.val)[::step]
                    for ii in h_line_indicies:
                        self.scan_v_positions[pixel_i] = self.v_array[jj]
                        self.scan_h_positions[pixel_i] = self.h_array[ii]
                        self.scan_index_array[pixel_i, :] = [kk, jj, ii]
                        pixel_i += 1
            for ii in range(self.Nh.val):
                self.scan_slow_move[pixel_i] = True
                for kk, step in [(2, 1), (3, -1)]:  # trace kk =2, retrace kk=3
                    v_line_indicies = range(self.Nv.val)[::step]
                    for jj in v_line_indicies:
                        self.scan_v_positions[pixel_i] = self.v_array[jj]
                        self.scan_h_positions[pixel_i] = self.h_array[ii]
                        self.scan_index_array[pixel_i, :] = [kk, jj, ii]
                        pixel_i += 1


class BaseRaster3DScan(Measurement):
    name = "base_raster_3D_scan"

    def __init__(
        self,
        app,
        h_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        v_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        z_limits: Tuple[float, float] = (-1_000_000_000_000, 1_000_000_000_000),
        h_unit: str = "",
        v_unit: str = "",
        z_unit: str = "",
        h_spinbox_decimals: int = 4,
        v_spinbox_decimals: int = 4,
        z_spinbox_decimals: int = 4,
        h_spinbox_step: float = 0.1,
        v_spinbox_step: float = 0.1,
        z_spinbox_step: float = 0.1,
        use_external_range_sync: bool = False,
        circ_roi_size: float = 1.0,
    ):
        self.h_spinbox_decimals = h_spinbox_decimals
        self.v_spinbox_decimals = v_spinbox_decimals
        self.z_spinbox_decimals = z_spinbox_decimals
        self.h_spinbox_step = h_spinbox_step
        self.v_spinbox_step = v_spinbox_step
        self.z_spinbox_step = z_spinbox_step
        self.h_limits = h_limits
        self.v_limits = v_limits
        self.z_limits = z_limits
        self.h_unit = h_unit
        self.v_unit = v_unit
        self.z_unit = z_unit
        self.use_external_range_sync = use_external_range_sync
        self.circ_roi_size = circ_roi_size
        Measurement.__init__(self, app)

    def setup(self):
        self.base_ui_filename = sibling_path(__file__, "raster_3d_scan_base.ui")
        self.ui: UI_Simple3DScanWindow = load_qt_ui_file(self.base_ui_filename)
        self.ui.h_axis_comboBox.setVisible(False)
        self.ui.v_axis_comboBox.setVisible(False)
        self.ui.z_axis_comboBox.setVisible(False)
        self.ui.setWindowTitle(self.name)

        self.display_update_period = 0.010  # seconds

        # connect events

        # local logged quantities
        h_lq_params = dict(
            vmin=self.h_limits[0],
            vmax=self.h_limits[1],
            unit=self.h_unit,
            spinbox_decimals=self.h_spinbox_decimals,
            spinbox_step=self.h_spinbox_step,
            dtype=float,
            ro=False,
        )
        h_range = self.h_limits[1] - self.h_limits[0]
        self.h0 = self.settings.New(
            "h0", initial=self.h_limits[0] + h_range * 0.25, **h_lq_params
        )
        self.h1 = self.settings.New(
            "h1", initial=self.h_limits[0] + h_range * 0.75, **h_lq_params
        )
        v_lq_params = dict(
            vmin=self.v_limits[0],
            vmax=self.v_limits[1],
            unit=self.v_unit,
            spinbox_decimals=self.v_spinbox_decimals,
            spinbox_step=self.v_spinbox_step,
            dtype=float,
            ro=False,
        )
        v_range = self.v_limits[1] - self.v_limits[0]
        self.v0 = self.settings.New(
            "v0", initial=self.v_limits[0] + v_range * 0.25, **v_lq_params
        )
        self.v1 = self.settings.New(
            "v1", initial=self.v_limits[0] + v_range * 0.75, **v_lq_params
        )

        z_lq_params = dict(
            vmin=self.z_limits[0],
            vmax=self.z_limits[1],
            unit=self.z_unit,
            spinbox_decimals=self.z_spinbox_decimals,
            spinbox_step=self.z_spinbox_step,
            dtype=float,
            ro=False,
        )
        z_range = self.z_limits[1] - self.z_limits[0]
        self.z0 = self.settings.New(
            "z0", initial=self.z_limits[0] + z_range * 0.25, **z_lq_params
        )
        self.z1 = self.settings.New(
            "z1", initial=self.z_limits[0] + z_range * 0.75, **z_lq_params
        )

        lq_params = dict(
            dtype=float, vmin=1e-9, vmax=abs(h_range), ro=False, unit=self.h_unit
        )
        self.dh = self.settings.New("dh", initial=self.h_spinbox_step, **lq_params)
        self.dh.spinbox_decimals = self.h_spinbox_decimals
        lq_params = dict(
            dtype=float, vmin=1e-9, vmax=abs(v_range), ro=False, unit=self.v_unit
        )
        self.dv = self.settings.New("dv", initial=self.v_spinbox_step, **lq_params)
        self.dv.spinbox_decimals = self.v_spinbox_decimals
        lq_params = dict(
            dtype=float, vmin=1e-9, vmax=abs(z_range), ro=False, unit=self.z_unit
        )
        self.dz = self.settings.New("dz", initial=self.z_spinbox_step, **lq_params)
        self.dz.spinbox_decimals = self.z_spinbox_decimals

        self.Nh = self.settings.New("Nh", initial=11, vmin=1, dtype=int, ro=False)
        self.Nv = self.settings.New("Nv", initial=11, vmin=1, dtype=int, ro=False)
        self.Nz = self.settings.New("Nz", initial=11, vmin=1, dtype=int, ro=False)

        self.h_center = self.settings.New(
            "h_center",
            dtype=float,
            ro=False,
            unit=self.h_unit,
            spinbox_decimals=self.h_spinbox_decimals,
            spinbox_step=self.h_spinbox_step,
        )
        self.v_center = self.settings.New(
            "v_center",
            dtype=float,
            ro=False,
            unit=self.v_unit,
            spinbox_decimals=self.v_spinbox_decimals,
            spinbox_step=self.v_spinbox_step,
        )
        self.z_center = self.settings.New(
            "z_center",
            dtype=float,
            ro=False,
            unit=self.z_unit,
            spinbox_decimals=self.z_spinbox_decimals,
            spinbox_step=self.z_spinbox_step,
        )

        self.h_span = self.settings.New(
            "h_span",
            dtype=float,
            ro=False,
            spinbox_decimals=self.h_spinbox_decimals,
            spinbox_step=self.h_spinbox_step,
            unit=self.h_unit,
        )
        self.v_span = self.settings.New(
            "v_span",
            dtype=float,
            ro=False,
            spinbox_decimals=self.v_spinbox_decimals,
            spinbox_step=self.v_spinbox_step,
            unit=self.v_unit,
        )
        self.z_span = self.settings.New(
            "z_span",
            dtype=float,
            ro=False,
            spinbox_decimals=self.z_spinbox_decimals,
            spinbox_step=self.z_spinbox_step,
            unit=self.z_unit,
        )

        self.Npixels = self.Nh.val * self.Nv.val * self.Nz.val

        self.scan_type = self.settings.New(
            "scan_type", dtype=str, initial="raster", choices=("raster", "serpentine")
        )

        self.continuous_scan = self.settings.New(
            "continuous_scan", dtype=bool, initial=False
        )
        self.settings.New("save_h5", dtype=bool, initial=True, ro=False)

        self.settings.New("show_previous_scans", dtype=bool, initial=True)

        self.settings.New(
            "pixel_time", dtype=float, ro=True, si=True, initial=1, unit="s"
        )
        self.settings.New("line_time", dtype=float, ro=True, si=True, unit="s")
        self.settings.New("frame_time", dtype=float, ro=True, si=True, unit="s")
        self.settings.New("total_time", dtype=float, ro=True, si=True, unit="s")

        for lq_name in ["Nh", "Nv", "Nz", "pixel_time"]:
            self.settings.get_lq(lq_name).add_listener(self.compute_times)

        self.compute_times()

        if not self.use_external_range_sync:
            self.h_range = LQRange(
                self.h0, self.h1, self.dh, self.Nh, self.h_center, self.h_span
            )
            self.v_range = LQRange(
                self.v0, self.v1, self.dv, self.Nv, self.v_center, self.v_span
            )
            self.z_range = LQRange(
                self.z0, self.z1, self.dz, self.Nz, self.z_center, self.z_span
            )

        for s in "h0 h1 dh v0 v1 dv z0 z1 dz".split():
            self.settings.get_lq(s).add_listener(self.compute_scan_params)

        self.scan_type.updated_value.connect(self.compute_scan_params)

        # connect events
        self.activation.connect_to_pushButton(self.ui.start_pushButton)

        self.h0.connect_to_widget(self.ui.h0_doubleSpinBox)
        self.h1.connect_to_widget(self.ui.h1_doubleSpinBox)
        self.v0.connect_to_widget(self.ui.v0_doubleSpinBox)
        self.v1.connect_to_widget(self.ui.v1_doubleSpinBox)
        self.z0.connect_to_widget(self.ui.z0_doubleSpinBox)
        self.z1.connect_to_widget(self.ui.z1_doubleSpinBox)
        self.dh.connect_to_widget(self.ui.dh_doubleSpinBox)
        self.dv.connect_to_widget(self.ui.dv_doubleSpinBox)
        self.dz.connect_to_widget(self.ui.dz_doubleSpinBox)
        self.Nh.connect_to_widget(self.ui.Nh_doubleSpinBox)
        self.Nv.connect_to_widget(self.ui.Nv_doubleSpinBox)
        self.Nz.connect_to_widget(self.ui.Nz_doubleSpinBox)
        self.scan_type.connect_to_widget(self.ui.scan_type_comboBox)

        self.progress.connect_to_widget(self.ui.progressBar)
        # self.progress.updated_value[str].connect(self.ui.xy_scan_progressBar.setValue)
        # self.progress.updated_value.connect(self.tree_progressBar.setValue)

        self.settings.get_lq("continuous_scan").connect_to_widget(
            self.ui.continuous_scan_checkBox
        )
        self.settings.get_lq("save_h5").connect_to_widget(self.ui.save_h5_checkBox)

        self.settings.get_lq("show_previous_scans").connect_to_widget(
            self.ui.show_previous_scans_checkBox
        )

        self.initial_scan_setup_plotting = False
        self.scan_specific_setup()

        self.add_operation("clear_previous_scans", self.clear_previous_scans)

        self.ui.clear_previous_scans_pushButton.clicked.connect(
            self.clear_previous_scans
        )

        self.compute_scan_params()

    def set_details_widget(self, widget=None, ui_filename=None):
        # print('LOADING DETAIL UI')
        if ui_filename is not None:
            details_ui = load_qt_ui_file(ui_filename)
        if widget is not None:
            details_ui = widget
        if hasattr(self, "details_ui"):
            if self.details_ui is not None:
                self.details_ui.deleteLater()
                self.ui.details_groupBox.layout().removeWidget(self.details_ui)
                # self.details_ui.hide()
                del self.details_ui
        self.details_ui = details_ui
        # return replace_widget_in_layout(self.ui.details_groupBox,details_ui)
        self.ui.details_groupBox.layout().addWidget(self.details_ui)
        return self.details_ui

    def set_h_limits(self, vmin, vmax, set_scan_to_max=False):
        self.settings.get_lq("h0").change_min_max(vmin, vmax)
        self.settings.get_lq("h1").change_min_max(vmin, vmax)
        if set_scan_to_max:
            self.settings["h0"] = vmin
            self.settings["h1"] = vmax

    def set_v_limits(self, vmin, vmax, set_scan_to_max=False):
        self.settings.get_lq("v0").change_min_max(vmin, vmax)
        self.settings.get_lq("v1").change_min_max(vmin, vmax)
        if set_scan_to_max:
            self.settings["v0"] = vmin
            self.settings["v1"] = vmax

    def set_z_limits(self, zmin, zmax, set_scan_to_max=False):
        self.settings.get_lq("z0").change_min_max(zmin, zmax)
        self.settings.get_lq("z1").change_min_max(zmin, zmax)
        if set_scan_to_max:
            self.settings["z0"] = zmin
            self.settings["z1"] = zmax

    def compute_scan_params(self):
        self.log.debug("compute_scan_params")
        # Don't recompute if a scan is running!
        if self.is_measuring():
            return  # maybe raise error

        # self.h_array = self.h_range.array #np.arange(self.h0.val, self.h1.val, self.dh.val, dtype=float)
        # self.v_array = self.v_range.array #np.arange(self.v0.val, self.v1.val, self.dv.val, dtype=float)

        # self.Nh.update_value(len(self.h_array))
        # self.Nv.update_value(len(self.v_array))

        self.range_extent = [self.h0.val, self.h1.val, self.v0.val, self.v1.val]

        # self.corners =  [self.h_array[0], self.h_array[-1], self.v_array[0], self.v_array[-1]]
        self.corners = self.range_extent

        self.imshow_extent = [
            self.h0.val - 0.5 * self.dh.val,
            self.h1.val + 0.5 * self.dh.val,
            self.v0.val - 0.5 * self.dv.val,
            self.v1.val + 0.5 * self.dv.val,
        ]

        self.compute_times()

        # call appropriate scan generator to determine scan size, don't compute scan arrays yet
        getattr(self, "gen_%s_scan" % self.scan_type.val)(gen_arrays=False)

    def compute_scan_arrays(self):
        print("params")
        self.compute_scan_params()
        gen_func_name = "gen_%s_scan" % self.scan_type.val
        print("gen_arrays:", gen_func_name)
        # calls correct scan generator function
        getattr(self, gen_func_name)(gen_arrays=True)

    def create_empty_scan_arrays(self):
        self.scan_h_positions = np.zeros(self.Npixels, dtype=float)
        self.scan_v_positions = np.zeros(self.Npixels, dtype=float)
        self.scan_z_positions = np.zeros(self.Npixels, dtype=float)
        self.scan_slow_move = np.zeros(self.Npixels, dtype=bool)
        self.scan_start_move = np.zeros(self.Npixels, dtype=bool)
        self.scan_index_array = np.zeros((self.Npixels, 3), dtype=int)

    def pre_run(self):
        # set all logged quantities read only
        for lqname in "h0 h1 v0 v1 z0 z1 dh dv dz Nh Nv Nz".split():
            self.settings.get_lq(lqname).change_readonly(True)

        self.compute_scan_params()

    def post_run(self):
        # set all logged quantities writable
        for lqname in "h0 h1 v0 v1 z0 z1 dh dv dz Nh Nv Nz".split():
            self.settings.get_lq(lqname).change_readonly(False)

    def clear_qt_attr(self, attr_name):
        if hasattr(self, attr_name):
            attr = getattr(self, attr_name)
            attr.deleteLater()
            del attr

    def setup_figure(self):
        self.compute_scan_params()

        self.clear_qt_attr("graph_layout")
        self.graph_layout = pg.GraphicsLayoutWidget(border=(100, 100, 100))
        self.ui.plot_groupBox.layout().addWidget(self.graph_layout)

        self.clear_qt_attr("img_plot")
        self.img_plot: pg.PlotItem = self.graph_layout.addPlot()

        self.img_items = []

        self.img_item = pg.ImageItem()
        self.img_items.append(self.img_item)

        self.img_plot.addItem(self.img_item)
        self.img_plot.showGrid(x=True, y=True)
        self.img_plot.setAspectLocked(lock=True, ratio=1)

        self.hist_lut = pg.HistogramLUTItem()
        self.graph_layout.addItem(self.hist_lut)

        # self.clear_qt_attr('current_stage_pos_arrow')
        self.current_stage_pos_arrow = pg.ArrowItem()
        self.current_stage_pos_arrow.setZValue(100)
        self.img_plot.addItem(self.current_stage_pos_arrow)

        self.graph_layout.nextRow()
        self.pos_label = pg.LabelItem(justify="right")
        self.pos_label.setText("=====")
        self.graph_layout.addItem(self.pos_label)

        self.scan_roi = pg.ROI([0, 0], [1, 1], movable=True)
        self.scan_roi.addScaleHandle([1, 1], [0, 0])
        self.scan_roi.addScaleHandle([0, 0], [1, 1])
        self.update_scan_roi()
        self.scan_roi.sigRegionChangeFinished.connect(self.mouse_update_scan_roi)

        self.img_plot.addItem(self.scan_roi)
        for lqname in "h0 h1 v0 v1 dh dv".split():
            self.settings.get_lq(lqname).updated_value.connect(self.update_scan_roi)

        self.img_plot.scene().sigMouseMoved.connect(self.mouseMoved)

        # GoTo position context menu
        # self.goto_cmenu_action = QtWidgets.QAction("GoTo Position", self.img_plot.scene())
        # self.img_plot.scene().contextMenu.append(self.goto_cmenu_action)
        # self.goto_cmenu_action.triggered.connect(self.on_goto_position)

        # Point ROI
        self.pt_roi = pg.CircleROI(
            (0, 0), (self.circ_roi_size, self.circ_roi_size), movable=True, pen=(0, 9)
        )
        # self.pt_roi.removeHandle(self.pt_roi.getHandles()[0])

        h = self.pt_roi.addTranslateHandle((0.5, 0.5))

        h.pen = pg.mkPen("r")
        h.update()
        self.img_plot.addItem(self.pt_roi)
        self.pt_roi.removeHandle(0)
        # self.pt_roi_plotline = pg.PlotCurveItem([0], pen=(0,9))
        # self.imview.getView().addItem(self.pt_roi_plotline)
        self.pt_roi.sigRegionChangeFinished[object].connect(self.on_update_pt_roi)

        self.connect_pos_widgets()

    def on_update_pt_roi(self, roi=None):
        if roi is None:
            roi = self.circ_roi
        roi_state = roi.saveState()
        x0, y0 = roi_state["pos"]
        xc = x0 + self.circ_roi_size / 2.0
        yc = y0 + self.circ_roi_size / 2.0
        self.new_pt_pos(xc, yc)

    def new_pt_pos(self, x, y):
        print("new_pt_pos", x, y)

    def mouse_update_scan_roi(self):
        x0, y0 = self.scan_roi.pos()
        w, h = self.scan_roi.size()
        self.h_center.update_value(x0 + w / 2)
        self.v_center.update_value(y0 + h / 2)
        self.h_span.update_value(w - self.dh.val)
        self.v_span.update_value(h - self.dv.val)
        self.compute_scan_params()
        self.update_scan_roi()

    def update_scan_roi(self):
        self.log.debug("update_scan_roi")
        x0, x1, y0, y1 = self.imshow_extent
        self.scan_roi.blockSignals(True)
        self.scan_roi.setPos((x0, y0, 0))
        self.scan_roi.setSize((x1 - x0, y1 - y0, 0))
        self.scan_roi.blockSignals(False)

    def disconnect_pos_widgets(self):
        if hasattr(self, "stage"):
            s: LQCollection = self.stage.settings
            s.get_lq("x_position").disconnect_from_widget(self.ui.x_doubleSpinBox)
            s.get_lq("y_position").disconnect_from_widget(self.ui.y_doubleSpinBox)
            s.get_lq("z_position").disconnect_from_widget(self.ui.y_doubleSpinBox)

    def connect_pos_widgets(self):
        self.disconnect_pos_widgets()

        if hasattr(self, "stage"):
            print(self.name, "connect_pos_widgets")
            s: LQCollection = self.stage.settings

            s.get_lq("x_position").updated_value.connect(self.update_arrow_pos)
            s.get_lq("y_position").updated_value.connect(self.update_arrow_pos)

            s.get_lq("x_position").connect_to_widget(self.ui.x_doubleSpinBox)
            s.get_lq("y_position").connect_to_widget(self.ui.y_doubleSpinBox)
            s.get_lq("z_position").connect_to_widget(self.ui.z_doubleSpinBox)

    def update_arrow_pos(self):
        if hasattr(self, "stage"):
            x = self.stage.settings["x_position"]
            y = self.stage.settings["y_position"]
            self.current_stage_pos_arrow.setPos(x, y)

    def on_goto_position(self):
        pass

    def update_display(self):
        # self.log.debug('update_display')
        if self.initial_scan_setup_plotting:
            if self.settings["show_previous_scans"]:
                self.img_item = pg.ImageItem()
                self.img_items.append(self.img_item)
                self.img_plot.addItem(self.img_item)
                self.hist_lut.setImageItem(self.img_item)

            self.img_item.setImage(self.display_image_map[0, :, :])
            x0, x1, y0, y1 = self.imshow_extent
            self.log.debug(f"update_display set bounds {x0} {x1} {y0} {y1}")
            self.img_item_rect = QtCore.QRectF(x0, y0, x1 - x0, y1 - y0)
            self.img_item.setRect(self.img_item_rect)
            self.log.debug(f"update_display set bounds {self.img_item_rect}")

            self.initial_scan_setup_plotting = False
        else:
            # if self.settings.scan_type.val in ['raster']
            kk, jj, ii = self.current_scan_index
            self.disp_img = self.display_image_map[kk, :, :].T
            self.img_item.setImage(self.disp_img, autoRange=False, autoLevels=True)
            # Important to set rectangle after setImage for non-square pixels
            self.img_item.setRect(self.img_item_rect)
            self.update_LUT()

    def update_LUT(self):
        """override this function to control display LUT scaling"""
        self.hist_lut.imageChanged(autoLevel=True)
        # DISABLE below because of crashing
        # non_zero_index = np.nonzero(self.disp_img)
        # if len(non_zero_index[0]) > 0:
        #     self.hist_lut.setLevels(*np.percentile(self.disp_img[non_zero_index],(1,99)))

    def clear_previous_scans(self):
        # current_img = img_items.pop()
        for img_item in self.img_items[:-1]:
            print("removing", img_item)
            self.img_plot.removeItem(img_item)
            img_item.deleteLater()

        self.img_items = [self.img_item]

    def mouseMoved(self, evt):
        mousePoint = self.img_plot.vb.mapSceneToView(evt)
        # print mousePoint

        # self.pos_label_text = "H {:+02.2f} um [{}], V {:+02.2f} um [{}]: {:1.2e} Hz ".format(
        #                mousePoint.x(), ii, mousePoint.y(), jj,
        #                self.count_rate_map[jj,ii]
        #                )

        self.pos_label.setText(
            f"H {mousePoint.x():+02.2f} um [{0}], V {mousePoint.y():+02.2f} um [{0}]: {0:1.2e} Hz"
        )

    def scan_specific_setup(self):
        "subclass this function to setup additional logged quantities and gui connections"
        pass
        # self.stage = self.app.hardware.dummy_xy_stage

        # self.app.hardware_components['dummy_xy_stage'].get_lq("x_position").connect_to_widget(self.ui.x_doubleSpinBox)
        # self.app.hardware_components['dummy_xy_stage'].get_lq("y_position").connect_to_widget(self.ui.y_doubleSpinBox)

        # self.app.hardware_components['apd_counter'].int_time.connect_to_widget(self.ui.int_time_doubleSpinBox)

        # logged quantities
        # connect events

    def initialize_controller(self):
        #        self.controller = self.app.hardware['xbox_controller']
        #
        #         if hasattr(self, 'controller'):
        #             self.pt_roi.sigRegionChangeFinished.connect(self.on_update_pt_roi)
        pass

    def update_point_roi_xbox(self):
        """Not yet implemented."""
        dx = self.controller.settings["Axis_4"]
        dy = self.controller.settings["Axis_3"]
        x, y = self.pt_roi.pos()
        if abs(dx) < 0.25:
            dx = 0
        if abs(dy) < 0.25:
            dy = 0
        if dx != 0 or dy != 0:
            c = self.controller.settings["sensitivity"]
            self.pt_roi.setPos(x + (c * dx), y + (c * dy))

    @property
    def h_array(self):
        # return self.h_range.array
        return np.linspace(self.h0.val, self.h1.val, self.Nh.val)

    @property
    def v_array(self):
        # return self.v_range.array
        return np.linspace(self.v0.val, self.v1.val, self.Nv.val)

    @property
    def z_array(self):
        # return self.z_range.array
        return np.linspace(self.z0.val, self.z1.val, self.Nz.val)

    def compute_times(self):
        # self.settings['pixel_time'] = 1.0/self.scanDAQ.settings['dac_rate']
        S = self.settings
        S["line_time"] = S["pixel_time"] * S["Nh"]
        S["frame_time"] = S["line_time"] * S["Nv"]
        S["total_time"] = S["frame_time"] * S["Nz"]

    #### Scan Generators
    def gen_raster_scan(self, gen_arrays=True):
        self.Npixels = self.Nh.val * self.Nv.val * self.Nz.val
        self.scan_shape = (self.Nz.val, self.Nv.val, self.Nh.val)

        if gen_arrays:
            # print "t0", time.time() - t0
            self.create_empty_scan_arrays()
            # print "t1", time.time() - t0

            #             t0 = time.time()
            #             pixel_i = 0
            #             for jj in range(self.Nv.val):
            #                 #print "tjj", jj, time.time() - t0
            #                 self.scan_slow_move[pixel_i] = True
            #                 for ii in range(self.Nh.val):
            #                     self.scan_v_positions[pixel_i] = self.v_array[jj]
            #                     self.scan_h_positions[pixel_i] = self.h_array[ii]
            #                     self.scan_index_array[pixel_i,:] = [0, jj, ii]
            #                     pixel_i += 1
            #             print "for loop raster gen", time.time() - t0

            t0 = time.time()

            V, Z, H = np.meshgrid(self.v_array, self.z_array, self.h_array)
            self.scan_h_positions[:] = H.flat
            self.scan_v_positions[:] = V.flat
            self.scan_z_positions[:] = Z.flat
            # print(self.scan_h_positions[0:self.Nh.val])
            # print(self.scan_v_positions[0:self.Nh.val*self.Nv.val])
            # print(self.scan_z_positions[0:self.Nh.val*self.Nv.val])

            self.scan_slow_move[:: self.Nh.val] = True
            self.scan_start_move[:: self.Nh.val * self.Nv.val] = True

            JJ, KK, II = np.meshgrid(
                np.arange(self.Nv.val), np.arange(self.Nz.val), np.arange(self.Nh.val)
            )
            self.scan_index_array[:, 0] = KK.flat
            self.scan_index_array[:, 1] = JJ.flat
            self.scan_index_array[:, 2] = II.flat
            # self.scan_v_positions
            print("array flatten raster gen", time.time() - t0)

    def gen_serpentine_scan(self, gen_arrays=True):
        self.Npixels = self.Nh.val * self.Nv.val * self.Nz.val
        self.scan_shape = (self.Nz.val, self.Nv.val, self.Nh.val)

        if gen_arrays:
            self.create_empty_scan_arrays()
            pixel_i = 0
            for kk in range(self.Nz.val):
                self.scan_start_move[pixel_i] = True
                for jj in range(self.Nv.val):
                    self.scan_slow_move[pixel_i] = True

                    if jj % 2:  # odd lines
                        h_line_indicies = range(self.Nh.val)[::-1]
                    else:  # even lines -- traverse in opposite direction
                        h_line_indicies = range(self.Nh.val)

                    for ii in h_line_indicies:
                        self.scan_z_positions[pixel_i] = self.z_array[kk]
                        self.scan_v_positions[pixel_i] = self.v_array[jj]
                        self.scan_h_positions[pixel_i] = self.h_array[ii]
                        self.scan_index_array[pixel_i, :] = [kk, jj, ii]
                        pixel_i += 1


#
#     def gen_trace_retrace_scan(self, gen_arrays=True):
#         self.Npixels = 2*self.Nh.val*self.Nv.val
#         self.scan_shape = (2, self.Nv.val, self.Nh.val)
#
#         if gen_arrays:
#             self.create_empty_scan_arrays()
#             pixel_i = 0
#             for jj in range(self.Nv.val):
#                 self.scan_slow_move[pixel_i] = True
#                 for kk, step in [(0,1),(1,-1)]: # trace kk =0, retrace kk=1
#                     h_line_indicies = range(self.Nh.val)[::step]
#                     for ii in h_line_indicies:
#                         self.scan_v_positions[pixel_i] = self.v_array[jj]
#                         self.scan_h_positions[pixel_i] = self.h_array[ii]
#                         self.scan_index_array[pixel_i,:] = [kk, jj, ii]
#                         pixel_i += 1
#
#     def gen_ortho_raster_scan(self, gen_arrays=True):
#         self.Npixels = 2*self.Nh.val*self.Nv.val
#         self.scan_shape = (2, self.Nv.val, self.Nh.val)
#
#         if gen_arrays:
#             self.create_empty_scan_arrays()
#             pixel_i = 0
#             for jj in range(self.Nv.val):
#                 self.scan_slow_move[pixel_i] = True
#                 for ii in range(self.Nh.val):
#                     self.scan_v_positions[pixel_i] = self.v_array[jj]
#                     self.scan_h_positions[pixel_i] = self.h_array[ii]
#                     self.scan_index_array[pixel_i,:] = [0, jj, ii]
#                     pixel_i += 1
#             for ii in range(self.Nh.val):
#                 self.scan_slow_move[pixel_i] = True
#                 for jj in range(self.Nv.val):
#                     self.scan_v_positions[pixel_i] = self.v_array[jj]
#                     self.scan_h_positions[pixel_i] = self.h_array[ii]
#                     self.scan_index_array[pixel_i,:] = [1, jj, ii]
#                     pixel_i += 1
#
#     def gen_ortho_trace_retrace_scan(self, gen_arrays=True):
#         print("gen_ortho_trace_retrace_scan")
#         self.Npixels = 4*len(self.h_array)*len(self.v_array)
#         self.scan_shape = (4, self.Nv.val, self.Nh.val)
#
#         if gen_arrays:
#             self.create_empty_scan_arrays()
#             pixel_i = 0
#             for jj in range(self.Nv.val):
#                 self.scan_slow_move[pixel_i] = True
#                 for kk, step in [(0,1),(1,-1)]: # trace kk =0, retrace kk=1
#                     h_line_indicies = range(self.Nh.val)[::step]
#                     for ii in h_line_indicies:
#                         self.scan_v_positions[pixel_i] = self.v_array[jj]
#                         self.scan_h_positions[pixel_i] = self.h_array[ii]
#                         self.scan_index_array[pixel_i,:] = [kk, jj, ii]
#                         pixel_i += 1
#             for ii in range(self.Nh.val):
#                 self.scan_slow_move[pixel_i] = True
#                 for kk, step in [(2,1),(3,-1)]: # trace kk =2, retrace kk=3
#                     v_line_indicies = range(self.Nv.val)[::step]
#                     for jj in v_line_indicies:
#                         self.scan_v_positions[pixel_i] = self.v_array[jj]
#                         self.scan_h_positions[pixel_i] = self.h_array[ii]
#                         self.scan_index_array[pixel_i,:] = [kk, jj, ii]
#                         pixel_i += 1
def ijk_zigzag_generator(dims, axis_order=(0, 1, 2)):
    """3D zig-zag scan pattern generator with arbitrary fast axis order"""

    ax0, ax1, ax2 = axis_order

    for i_ax0 in range(dims[ax0]):
        zig_or_zag0 = (1, -1)[i_ax0 % 2]
        for i_ax1 in range(dims[ax1])[::zig_or_zag0]:
            zig_or_zag1 = (1, -1)[(i_ax0 + i_ax1) % 2]
            for i_ax2 in range(dims[ax2])[::zig_or_zag1]:

                ijk = [0, 0, 0]
                ijk[ax0] = i_ax0
                ijk[ax1] = i_ax1
                ijk[ax2] = i_ax2

                yield tuple(ijk)
    return



from pathlib import Path
import unittest

import numpy as np

from ScopeFoundry import BaseMicroscopeApp, HardwareComponent, Measurement
from ScopeFoundry.base_app.base_app import WRITE_RES


class Measure1(Measurement):
    name = "measure1"

    def setup(self):
        self.settings.New("string", str, ro=True, initial="0")
        self.settings.New("float", float, ro=True, initial=0.0)
        self.settings.New("choices", int, choices=INITIAL_CHOICES)
        self.settings.New("array", bool, is_array=True, initial=[False, False, False])


class Hardware1(HardwareComponent):
    name = "hardware1"

    def setup(self):
        self.settings.New("string", str, ro=True, initial="0")
        self.settings.New("float", float, ro=True, initial=0.0)
        self.settings.New("choices", int, choices=INITIAL_CHOICES)
        self.settings.New("protected_int", int, initial=0, protected=True)


INITIAL_CHOICES = [("choice 0", 0), ("choice 1", 1), ("choice 2", 2), ("choice 3", 3)]


class SettingsIOTest(unittest.TestCase):

    def setUp(self):
        self.app = BaseMicroscopeApp([])
        self.ms = self.app.add_measurement(Measure1(self.app))
        self.hw = self.app.add_hardware(Hardware1(self.app))
        self.root = Path(__file__).parent

        self.mms = self.ms.settings
        self.hws = self.hw.settings

    def test_initial_value_set(self):
        self.assertEqual(self.mms["string"], "0")
        self.assertEqual(self.mms["float"], 0.0)
        self.assertEqual(self.mms["choices"], 0)
        self.assertTrue(np.all(self.mms["array"] == np.array([False, False, False])))
        self.assertEqual(self.hws["string"], "0")
        self.assertEqual(self.hws["float"], 0.0)
        self.assertEqual(self.hws["choices"], 0)

    def test_all_correct(self):
        self.app.settings_load_ini(
            self.root / "settings_io_test_correct.ini",
            show_report=False,
        )
        self.assertEqual(self.mms["string"], "1")
        self.assertEqual(self.mms["float"], 1.0)
        self.assertEqual(self.mms["choices"], 1)
        self.assertEqual(self.hws["string"], "1")
        self.assertEqual(self.hws["float"], 1.0)
        self.assertEqual(self.hws["choices"], 1)
        self.assertEqual(self.hws["protected_int"], 0)

    def test_first_section_wrong_and_continue(self):
        self.app.settings_load_ini(
            self.root / "settings_io_test_measure1_false.ini", show_report=False
        )
        self.assertEqual(self.hws["string"], "2")
        self.assertEqual(self.hws["float"], 2.0)
        self.assertEqual(self.hws["choices"], 2)

    def test_reporting(self):
        self.app.settings_load_ini(
            self.root / "settings_io_test_measure1_false_2.ini", show_report=False
        )
        self.assertTrue("meeeeeeasurement/measure1/straaaaang" in self.app._report)
        self.assertTrue(
            self.app._report["meeeeeeasurement/measure1/straaaaang"]
            is WRITE_RES.MISSING
        )
        self.assertTrue(
            self.app._report["hardware/hardware1/protected_int"], "PROTECTED"
        )

    def test_first_section_wrong_and_continue(self):
        fname = self.root / "settings_io_test_measure1_false_2.ini"
        self.app.settings_load_ini(fname, show_report=False)
        self.assertTrue("meeeeeeasurement/measure1/straaaaang" in self.app._report)
        self.assertTrue(
            self.app._report["meeeeeeasurement/measure1/straaaaang"]
            is WRITE_RES.MISSING
        )

        self.assertEqual(self.hws["string"], "2")
        self.assertEqual(self.hws["float"], 2.0)
        self.assertEqual(self.hws["choices"], 2)

    def test_first_setting_misspelled_and_continue(self):
        self.app.settings_load_ini(
            self.root / "settings_io_test_measure1_string_false.ini",
            show_report=False,
        )
        # self.assertEqual(self.hw1["string"], "2")
        self.assertEqual(self.hws["float"], 2.0)
        self.assertEqual(self.hws["choices"], 2)

    def test_roundtrip(self):

        # all values are initial values
        self.app.settings_save_ini(self.root / "settings_io_test_roundtrip.ini")

        # change them up
        self.mms["string"] = "1"
        self.mms["float"] = 1.0
        self.mms["choices"] = 1
        self.mms["array"] = np.array([False, False, True])
        self.hws["string"] = "1"
        self.hws["float"] = 1.0
        self.hws["choices"] = 1

        # make sure they are changed
        self.assertFalse(np.all(self.mms["array"] == np.array([False, False, False])))

        # load values
        self.app.settings_load_ini(
            self.root / "settings_io_test_roundtrip.ini", show_report=False
        )
        self.assertEqual(self.mms["string"], "0")
        self.assertEqual(self.mms["float"], 0.0)
        self.assertEqual(self.mms["choices"], 0)
        self.assertTrue(np.all(self.mms["array"] == np.array([False, False, False])))
        self.assertEqual(self.hws["string"], "0")
        self.assertEqual(self.hws["float"], 0.0)
        self.assertEqual(self.hws["choices"], 0)


if __name__ == "__main__":
    unittest.main()



# generated with ScopeFoundry.analyze_with_ipynb()
# Probality you want to
#
# from h5_data_loaders import load
# data = load('your_file_name.h5')

import functools
from dataclasses import dataclass
from pathlib import Path

import h5py
import numpy as np

load_funcs = {}


def get_measurement_name(fname):
    with h5py.File(fname) as file:
        if len(file["measurement"].keys()) == 1:
            return list(file["measurement"].keys())[0]
        return file.attrs["measurement"]


def load(fname: str):
    mm_name = get_measurement_name(fname)
    return load_funcs[mm_name](fname)


def load_settings(fname):
    path = Path(fname)
    if not path.suffix == ".h5":
        return {}

    settings = {}
    visit_func = functools.partial(_settings_visitfunc, settings=settings)

    with h5py.File(fname) as file:
        file.visititems(visit_func)

    return settings


def _settings_visitfunc(name, node, settings):
    if not name.endswith("settings"):
        return

    for key, val in node.attrs.items():
        lq_path = f"{name.replace('settings', key)}"
        settings[lq_path] = val

def get_mm_name(fname):
    with h5py.File(fname) as file:
        if len(file["measurement"].keys()) == 1:
            return list(file["measurement"].keys())[0]
        return file.attrs["measurement"]


@dataclass
class Example2DScan:
    settings:dict
    pixel_time: np.ndarray
    range_extent: np.ndarray
    corners: np.ndarray
    signal_map: np.ndarray
    scan_v_positions: np.ndarray
    h_array: np.ndarray
    imshow_extent: np.ndarray
    v_array: np.ndarray
    scan_h_positions: np.ndarray
    scan_slow_move: np.ndarray
    scan_index_array: np.ndarray


def load_example_2d_scan(fname:str) -> Example2DScan:
    with h5py.File(fname) as file:
        m = file['measurement/example_2d_scan']
        return Example2DScan(
            settings=load_settings(fname),
            pixel_time=m['pixel_time'][:] if 'pixel_time' in m else None,
            range_extent=m['range_extent'][:] if 'range_extent' in m else None,
            corners=m['corners'][:] if 'corners' in m else None,
            signal_map=m['signal_map'][:] if 'signal_map' in m else None,
            scan_v_positions=m['scan_v_positions'][:] if 'scan_v_positions' in m else None,
            h_array=m['h_array'][:] if 'h_array' in m else None,
            imshow_extent=m['imshow_extent'][:] if 'imshow_extent' in m else None,
            v_array=m['v_array'][:] if 'v_array' in m else None,
            scan_h_positions=m['scan_h_positions'][:] if 'scan_h_positions' in m else None,
            scan_slow_move=m['scan_slow_move'][:] if 'scan_slow_move' in m else None,
            scan_index_array=m['scan_index_array'][:] if 'scan_index_array' in m else None,
        )


load_funcs['example_2d_scan'] = load_example_2d_scan


@dataclass
class Example3DScan:
    settings:dict
    pixel_time: np.ndarray
    range_extent: np.ndarray
    corners: np.ndarray
    signal_map: np.ndarray
    scan_v_positions: np.ndarray
    h_array: np.ndarray
    imshow_extent: np.ndarray
    v_array: np.ndarray
    scan_h_positions: np.ndarray
    scan_slow_move: np.ndarray
    z_array: np.ndarray
    scan_z_positions: np.ndarray
    scan_index_array: np.ndarray


def load_example_3d_scan(fname:str) -> Example3DScan:
    with h5py.File(fname) as file:
        m = file['measurement/example_3d_scan']
        return Example3DScan(
            settings=load_settings(fname),
            pixel_time=m['pixel_time'][:] if 'pixel_time' in m else None,
            range_extent=m['range_extent'][:] if 'range_extent' in m else None,
            corners=m['corners'][:] if 'corners' in m else None,
            signal_map=m['signal_map'][:] if 'signal_map' in m else None,
            scan_v_positions=m['scan_v_positions'][:] if 'scan_v_positions' in m else None,
            h_array=m['h_array'][:] if 'h_array' in m else None,
            imshow_extent=m['imshow_extent'][:] if 'imshow_extent' in m else None,
            v_array=m['v_array'][:] if 'v_array' in m else None,
            scan_h_positions=m['scan_h_positions'][:] if 'scan_h_positions' in m else None,
            scan_slow_move=m['scan_slow_move'][:] if 'scan_slow_move' in m else None,
            z_array=m['z_array'][:] if 'z_array' in m else None,
            scan_z_positions=m['scan_z_positions'][:] if 'scan_z_positions' in m else None,
            scan_index_array=m['scan_index_array'][:] if 'scan_index_array' in m else None,
        )


load_funcs['example_3d_scan'] = load_example_3d_scan



import unittest

from ScopeFoundry.tests.unittests.test_get_settings import GetSettingsTest
from ScopeFoundry.tests.unittests.test_lq_choice_list import ChoiceListTest
from ScopeFoundry.tests.unittests.test_settings_io import SettingsIOTest
from ScopeFoundry.tests.unittests.test_lq_connection import LQConnectionTest
from ScopeFoundry.tests.unittests.test_lq_range import LQRangeTest
from ScopeFoundry.tests.unittests.test_analyze_nb import AnalyzeNBTest
from ScopeFoundry.tests.unittests.test_operations import TestOperations


# following also require visual inspection - run individual files
from ScopeFoundry.tests.hw_connect_failure_test import AppTest
from ScopeFoundry.tests.nested_measurement_test import NestMeasureTestAppTest
from ScopeFoundry.tests.quickbar_test import QuickbarTest

if __name__ == "__main__":
    unittest.main()



import sys

from ScopeFoundry import BaseMicroscopeApp
from ScopeFoundry.examples.ScopeFoundryHW.bsinc_noiser200 import Noiser200HW
from ScopeFoundry.examples.ScopeFoundryHW.simulon_xyz_stage import SimulonXYZStageHW
from ScopeFoundry.examples.measurements.example_2d_slowscan_measure import (
    Example2DSlowScanMeasure,
)
from ScopeFoundry.examples.measurements.example_3d_slowscan_measure import (
    Example3DSlowScanMeasure,
)


class FancyApp(BaseMicroscopeApp):

    name = "fancy app"

    def setup(self):

        self.add_hardware(SimulonXYZStageHW(self))
        self.add_hardware(Noiser200HW(self))

        # ## Example scan measurements
        # Each actuator can be defined with a tuple of settings paths. The following formats are supported:
        # 1. (name, position_path | position_read_func, target_position_path | target_position_write_func)
        # 2. (name, target_position_path | target_position_write_func) -> target_position_path=position_path
        # 3. (target_position_path) -> name=target_position_path=position_path
        actuators = (
            (
                "x_position",
                "hw/simulon_xyz_stage/x_position",
                "hw/simulon_xyz_stage/x_target_position",
            ),
            (
                "y_position",
                "hw/simulon_xyz_stage/y_position",
                "hw/simulon_xyz_stage/y_target_position",
            ),
            (
                "z_position",
                "hw/simulon_xyz_stage/z_position",
                "hw/simulon_xyz_stage/z_target_position",
            ),
        )

        self.add_measurement(Example2DSlowScanMeasure(self, actuators=actuators))
        self.add_measurement(Example3DSlowScanMeasure(self, actuators=actuators))

        # ## Other Built-in Measurements
        # https://scopefoundry.org/docs/20_built-in-measurements/
        #
        # from ScopeFoundry.sequencer import Sequencer, SweepSequencer

        # self.add_measurement(Sequencer)
        # self.add_measurement(SweepSequencer)

        # from ScopeFoundry import RangedOptimization

        # self.add_measurement(RangedOptimization(self))

        # from ScopeFoundry import PIDFeedbackControl

        # self.add_measurement(PIDFeedbackControl(self))


if __name__ == "__main__":
    app = FancyApp(sys.argv)
    app.settings_load_ini("default_settings.ini")
    sys.exit(app.exec_())



import sys


def invalid_input():
    print("----------------------------------------------------------")
    print("Invalid command. Please use one of the following commands:")
    print("----------------------------------------------------------")
    print_cmds()


def print_cmds():
    print("python -m ScopeFoundry init")
    print("python -m ScopeFoundry ipynb")
    print("python -m ScopeFoundry tools")
    print("python -m ScopeFoundry new_hardware")
    print("python -m ScopeFoundry new_measurement")
    print("python -m ScopeFoundry publish_hardware")


if __name__ == "__main__":

    cmd = sys.argv[1].lower() if len(sys.argv) > 1 else None
    if cmd is None:
        invalid_input()

    if cmd in ("init", "new_app"):
        from ScopeFoundry.tools.features.new_app import new_app

        run_str = new_app()

        print("")
        print("Or use one of the following commmands to extent your app")
        print("----------------------------------------------------------")
        print_cmds()
        print("----------------------------------------------------------")
        print("")
        print("To run your app, use the following command:")
        print(run_str)

    elif cmd in ("ipynb",):
        from ScopeFoundry.h5_analyze_with_ipynb import analyze_with_ipynb

        analyze_with_ipynb()
    elif cmd in ("tools",):
        from ScopeFoundry.tools.app import start_app

        start_app()
    elif cmd in ("new_hw", "new_hardware"):
        from ScopeFoundry.tools.app import start_app

        start_app("new hardware")
    elif cmd in ("new_mm", "new_measurement"):
        from ScopeFoundry.tools.features.new_measurement import main

        main()

    elif cmd in ("publish_hw", "publish_hardware"):
        from ScopeFoundry.tools.app import start_app

        start_app("publish HW on GitHub")

    else:
        invalid_input()



"""
Created on Jul 20, 2018

@author: Benedikt Ursprung
"""

from qtpy import QtCore, QtWidgets, QtGui
import pyqtgraph as pg
import numpy as np


class MinMaxQSlider(QtWidgets.QWidget):
    """
    Costume QSlider widget designed to work with logged_quantities.
    Two spin boxes allow to dynamically update the min and max value of the slider.
    """

    # Signals
    updated_value = QtCore.Signal((float,), (int,))

    def __init__(
        self,
        name="",
        unit="",
        initial=50,
        vmin=0,
        vmax=100,
        spinbox_decimals=2,
        **QSlider_kwargs,
    ):

        self.SLIDER_MIN = -2147483648
        self.SLIDER_MAX = 2147483647

        self.name = name
        self.unit = unit

        # Setup widget
        QtWidgets.QWidget.__init__(self)

        # slider
        self.slider = QtWidgets.QSlider(
            orientation=QtCore.Qt.Horizontal, **QSlider_kwargs
        )
        self.slider.valueChanged.connect(self.on_slider_value_changed)
        self.slider.valueChanged.connect(self.set_title)

        # min and max spinbox
        self.min_double_spinbox = QtWidgets.QSpinBox()
        self.max_double_spinbox = QtWidgets.QSpinBox()
        self.min_double_spinbox.setRange(self.SLIDER_MIN, self.SLIDER_MAX)
        self.max_double_spinbox.setRange(self.SLIDER_MIN, self.SLIDER_MAX)

        self.set_range(vmin, vmax)

        self.min_double_spinbox.valueChanged[int].connect(self.update_slider_minimum)
        self.max_double_spinbox.valueChanged[int].connect(self.update_slider_maximum)

        self.spinbox_decimals = spinbox_decimals

        # title label
        self.title_label = QtWidgets.QLabel()

        # Setup layout
        self.layout_ = QtWidgets.QGridLayout()
        self.layout_.addWidget(
            self.min_double_spinbox, 0, 0, alignment=QtCore.Qt.AlignLeft
        )
        self.layout_.addWidget(self.title_label, 0, 1, alignment=QtCore.Qt.AlignHCenter)
        self.layout_.addWidget(
            self.max_double_spinbox, 0, 2, alignment=QtCore.Qt.AlignRight
        )
        self.layout_.addWidget(self.slider, 1, 0, 1, 3)

        self.setLayout(self.layout_)
        self.update_value(initial)

    def set_range(self, vmin, vmax):
        val = self.val

        vmin = min(max(vmin, self.SLIDER_MIN), self.SLIDER_MAX)
        vmax = max(min(vmax, self.SLIDER_MAX), self.SLIDER_MIN)

        self.min_double_spinbox.setValue(vmin)
        self.max_double_spinbox.setValue(vmax)

        self.slider.blockSignals(True)
        self.slider.setRange(int(vmin), int(vmax))
        self.slider.blockSignals(False)

        self.update_value(val)

    def set_name(self, name):
        self.name = name
        self.set_title(self.val)

    def value(self):
        return self.val

    @property
    def val(self):
        return self.transform_from_slider(self.slider.value())

    @property
    def vmin(self):
        return self.min_double_spinbox.value()

    @property
    def vmax(self):
        return self.max_double_spinbox.value()

    @property
    def vrange(self):
        return self.vmax - self.vmin

    def transform_to_slider(self, x):
        return int(x)

    def transform_from_slider(self, x):
        return x

    def update_slider_minimum(self, vmin):
        self.set_range(vmin, self.vmax)

    def update_slider_maximum(self, vmax):
        self.set_range(self.vmin, vmax)

    def update_value(self, x):
        self.slider.setValue(self.transform_to_slider(x))

    def on_slider_value_changed(self, x):
        val = self.transform_from_slider(self.slider.value())
        self.updated_value.emit(val)

    def set_title(self, val):
        val = self.transform_from_slider(val)
        self.title_label.setText(f"{self.name} {self.val} {self.unit}")


class RegionSlicer(QtWidgets.QWidget):
    """
    **Bases:** :class: `QWidget <pyqt.QtWidgets.QWidget>`

    Adds a movable <pyqtgraph.LinearRegionItem> to a plot.
    Provides a numpy slice and mask that would slices the x_array according to the region.

    ===============================  =============================================================================
    **Signals:**
    region_changed_signal()          Emitted when region is changed or activated by user
    ===============================  =============================================================================


    ===============================  =============================================================================
    **Properties:**
    slice:                           numpy.slice[start:stop]
    s_:                              :property:slice if activated  else *s_return_on_deactivated*
    mask:                            numpy boolean array of size 'x_array': True for where x is within region
                                     false otherwise (ignores step). Useful if x_array is not ordered.
    m_:                              mask if activated, else a mask equivalent to
                                        *s_return_on_deactivated*
    ===============================  =============================================================================
    """

    region_changed_signal = QtCore.Signal()

    def __init__(
        self,
        plot_item,
        x_array=None,
        name="array_slicer_name",
        slicer_updated_func=lambda: None,
        initial=[0, 100],
        s_return_on_deactivated=np.s_[:],
        brush=QtGui.QColor(0, 255, 0, 70),
        ZValue=10,
        font=QtGui.QFont("Times", 12),
        label_line=1,
        activated=False,
    ):
        """Create a new LinearRegionItem on plot_item.

        ====================== ==============================================================
        **Arguments:**
        plot_item              <pyqtgraph.PlotDataItem> (recommended)
                                or <pyqtgraph.PlotItem>  (does not grab x_array data from plot
                                item: initialize x_array manually
                                or use :func:`set_x_array <self.set_x_array>`)
        x_array                use to initialize x_array if plot_item == <pyqtgraph.PlotItem>.
        name                   <str>
        slicer_updated_func    gets called when region is updated, alternatively use
                               :sig:region_changed_signal().
        initial                initial index values [start,stop,step] or [start,stop]
        s_return_on_deactivated  Object returned if RegionSlicer is not activated.
                                 Slicing with np.s_[:] (default) and np.s_[0] gives the full and
                                 an empty array respectively. Note, <RegionSlicer.slice> always
                                 returns the last determined slice even if slice is deactivated.
        brush,ZValue,          are passed to the LinearRegionItem
        font,                  passed to the label
        label_line             '0' or '1' for placement of label onto 'left' or 'right'
                               bounding line.
        activated              <bool> state at initialization
        ====================== ===============================================================
        """
        QtWidgets.QWidget.__init__(self)

        self.name = name

        self.slicer_updated = slicer_updated_func

        from ScopeFoundry.logged_quantity import LQCollection

        self.settings = LQCollection()
        if len(initial) == 2:
            initial.append(1)
        start, stop, step = initial
        self.start = self.settings.New("start", int, initial=start, vmin=0)
        self.stop = self.settings.New("stop", int, initial=stop, vmin=0)
        self.step = self.settings.New("step", int, initial=step)
        self.activated = self.settings.New("activated", bool, initial=activated)

        self.s_return_on_deactivated = s_return_on_deactivated

        if type(plot_item) == pg.PlotDataItem:
            self.plot_data_item = plot_item
            self.plot_data_item.sigPlotChanged.connect(self.set_x_array_from_data_item)
            self.parent_plot_item = plot_item.parentItem()
        elif type(plot_item) == pg.PlotItem:
            self.plot_data_item = None
            self.parent_plot_item = plot_item

        self.linear_region_item = pg.LinearRegionItem(brush=brush)
        self.linear_region_item.setZValue(ZValue)
        self.parent_plot_item.addItem(self.linear_region_item)

        self.inf_line_label = pg.InfLineLabel(
            self.linear_region_item.lines[label_line],
            self.name,
            position=0.78,
            anchor=(0.5, 0.5),
        )
        # self.inf_line_label = pg.TextItem(self.name, anchor=(0.5, 0.5))
        self.inf_line_label.setFont(font)
        self.set_label("")

        if x_array == None:  # give it something to work with.
            x_array = np.arange(512)
        self.set_x_array(x_array)

        self.start.add_listener(self.on_change_start_stop)
        self.stop.add_listener(self.on_change_start_stop)
        self.activated.add_listener(self.on_change_activated)
        self.linear_region_item.sigRegionChangeFinished.connect(self.on_change_region)

    @QtCore.Slot(object)
    def set_x_array_from_data_item(self):
        x_array, _ = self.plot_data_item.getData()
        if not np.array_equal(x_array, self.x_array):
            self.apply_new_x_array(x_array)
            # print('set_x_array_from_data_item: new x_array()')
        else:
            pass

    def set_x_array(self, x_array):
        """
        use this function to update the x_array
        not required to use if type(plot_item) == pg.PlotDataItem
        """
        self.apply_new_x_array(x_array)

    def apply_new_x_array(self, x_array):
        mn, mx = self.linear_region_item.getRegion()
        self.x_array = x_array
        self.linear_region_item.setBounds([self.x_array.min(), self.x_array.max()])
        self.settings["start"] = np.argmin((x_array - mn) ** 2)
        self.settings["stop"] = np.argmin((x_array - mx) ** 2)

    @property
    def slice(self):
        S = self.settings
        return np.s_[S["start"] : S["stop"] : S["step"]]

    @property
    def s_(self):
        """returns slice based on region if activated else `s_return_on_deactivated`"""
        if self.activated.val:
            return self.slice
        else:
            return self.s_return_on_deactivated

    @property
    def mask(self):
        Xmin = self.x_array[self.settings["start"]]
        Xmax = self.x_array[self.settings["stop"]]
        return (self.x_array >= Xmin) * (self.x_array <= Xmax)

    @property
    def m_(self):
        if self.activated.val:
            return self.mask
        else:
            s = self.s_return_on_deactivated
            Xmin = self.x_array[s.start]
            Xmax = self.x_array[s.stop]
            return (self.x_array >= Xmin) * (self.x_array <= Xmax)

    @QtCore.Slot(object)
    def on_change_region(self):
        """
        updates settings based on region
        """
        print(self.name, "on_change_region")
        self.region_min, self.region_max = mn, mx = self.linear_region_item.getRegion()
        self.settings["start"] = np.argmin((self.x_array - mn) ** 2)
        self.settings["stop"] = np.argmin((self.x_array - mx) ** 2)
        self.region_changed_signal.emit()
        self.slicer_updated()

    def on_change_start_stop(self):
        """
        updates linear_region_item based on settings
        """
        rgn = [0, 0]
        rgn[0] = self.x_array[self.settings["start"]]
        try:
            rgn[1] = self.x_array[self.settings["stop"]]
        except IndexError:
            rgn[1] = self.x_array[self.settings["start"]]
        self.linear_region_item.setRegion(rgn)

    def on_change_activated(self):
        activated = self.activated.val
        self.linear_region_item.setEnabled(activated)
        self.linear_region_item.setAcceptHoverEvents(activated)
        self.linear_region_item.setAcceptTouchEvents(activated)
        self.linear_region_item.setVisible(activated)
        self.region_changed_signal.emit()
        self.slicer_updated()

    def New_UI(self, include=None, exclude=[]):
        ui_widget = self.settings.New_UI(include, exclude)
        ui_widget.layout().insertRow(0, QtWidgets.QLabel(f"<b>{self.name}</b>"))
        return ui_widget

    def set_label(self, text="", title=None, color=(200, 200, 200)):
        if title == None:
            title = self.name
        self.inf_line_label.setHtml(f"<h3>{title}</h3>{text}")
        self.inf_line_label.setColor(color)

    def get_sliced_xar(self, ar=None):
        if ar == None:
            if hasattr(self, "plot_data_item"):
                _, y = self.plot_data_item.getData()
                return (self.x_array[self.s_], y[self.s_])
            else:
                return None
        else:
            return (self.x_array[self.s_], ar[self.s_])



# required for Sweep4d example


from ScopeFoundry import Collector


class Noiser200Collector(Collector):
    name = "noiser200"
    repeated_dset_names = ("signals",)
    acquisition_duration_path = "mm/noiser200/capture_duration"

    def run(
        self,
        index,
        host_measurement,
        polling_func=None,
        polling_time=0.001,
        int_time=None,
        *args,
        **kwargs
    ):
        # needs to define self.data with reapated_dset_names
        self.data["signals"] = self.app.get_lq(
            "hw/noiser_200/signal"
        ).read_from_hardware()



import subprocess
from collections import deque
from enum import Enum
from functools import partial
from inspect import signature
from typing import Any, Iterable, List, Tuple, Union

import pyqtgraph as pg
from qtpy import QtCore, QtGui, QtWidgets

from ScopeFoundry.helper_funcs import QLock, bool2str, get_logger_from_class, str2bool
from ScopeFoundry.widgets import MinMaxQSlider


class LoggedQuantity(QtCore.QObject):
    """
    **LoggedQuantity** objects are containers that wrap settings. These settings
    may be a number (integer or float) or a string and occasionally small arrays
    of them.

    These objects emit signals when changed and can be connected bidirectionally
    to Qt Widgets.

    In ScopeFoundry we represent the values in an object called a
    `LoggedQuantity`. A :class:`LoggedQuantity` is a class that contains a
    value, a `bool`, `float`, `int`, `str` etc that is part of an application's
    state. In the case of microscope and equipment control, these also can
    represent the state of a piece of hardware. These are very useful objects
    because the are the central location of the value contained within. All
    graphical interface views will be guaranteed to be consistent with the `LQ`
    state. The data of these quantities will also be saved in datafiles created
    by ScopeFoundry.

    """

    # signal sent when value has been updated
    updated_value = QtCore.Signal((float,), (int,), (bool,), (), (str,))
    # signal sent when value has been updated, sends text representation
    updated_text_value = QtCore.Signal(str)
    # emits the index of the value in self.choices
    updated_choice_index_value = QtCore.Signal(int)

    # signal sent when min max range updated
    updated_min_max = QtCore.Signal((float, float), (int, int), ())
    # signal sent when read only (ro) status has changed
    updated_readonly = QtCore.Signal((bool,), ())

    def __init__(
        self,
        name,
        dtype=float,
        initial=0,
        fmt: str = "%g",
        si: bool = False,
        ro: bool = False,  # read only flag
        unit: str = None,
        spinbox_decimals: int = 2,
        spinbox_step: float = 0.1,
        vmin: float = -1e12,
        vmax: float = +1e12,
        choices: Iterable = None,
        reread_from_hardware_after_write: bool = False,
        description: str = None,
        colors=None,
        protected: bool = False,  # a guard that prevents from being updated, i.e. file loading
        is_cmd: bool = False,
        is_clipboardable: bool = False,
        default_widget_factory=None,
    ):
        QtCore.QObject.__init__(self)

        self.name = name

        if dtype in [int, str, float]:
            pass
        elif dtype in ["int", "uint"]:
            dtype = int
        elif dtype in ["float", "float32"]:
            dtype = float

        self.dtype = dtype

        # choices should be tuple [ ('name', val) ... ] or simple list [val, val, ...]
        self.choices = self._expand_choices(choices)
        if self.choices and not initial in (x[1] for x in self.choices):
            initial = self.choices[0][1]

        self.val = dtype(initial)
        self.hardware_read_func = None
        self.hardware_set_func = None
        self.fmt = fmt  # string formatting string. This is ignored if dtype==str
        if self.dtype == str:
            self.fmt = "%s"
        self.si = si  # will use pyqtgraph SI Spinbox if True
        self.unit = unit
        self.vmin = vmin
        self.vmax = vmax

        # self.change_choice_list(choices)
        self.ro = ro  # Read-Only
        self.is_array = False
        self.description = description

        self.colors = colors
        if colors:
            self.qcolors = [to_q_color(color) for color in colors]
        else:
            self.qcolors = []

        self.log = get_logger_from_class(self)

        if self.dtype == int:
            self.spinbox_decimals = 0
        else:
            self.spinbox_decimals = spinbox_decimals
        self.reread_from_hardware_after_write = reread_from_hardware_after_write

        if self.dtype == int:
            self.spinbox_step = 1
        else:
            self.spinbox_step = spinbox_step

        self.default_widget_factory = default_widget_factory

        self.oldval = None

        self._in_reread_loop = False  # flag to prevent reread from hardware loops

        self.widget_list = []
        self.listeners = []

        # threading lock
        self.lock = QLock(mode=1)  # mode 0 is non-reentrant lock

        self.path = ""

        self.protected = protected

        self.is_clipboardable = is_clipboardable
        self.is_cmd = is_cmd

        self.prev_vals = deque([], 3)
        self.proposed_values = deque([], 7)

    def coerce_to_type(self, x):
        """
        Force x to dtype of the LQ

        =============  ==================================
        **Arguments**  **Description**
        *x*            value of type str, bool, int, etc.
        =============  ==================================

        :returns: Same value, *x* of the same type as its respective logged
        quantity

        """
        if self.dtype == bool and isinstance(x, str):
            return str2bool(x)
        return self.dtype(x)

    def coerce_to_str(self, x):
        if self.dtype == bool:
            return bool2str(x)
        else:
            return str(x)

    @property
    def val_str(self):
        self.coerce_to_str(self.val)

    def _expand_choices(self, choices: Iterable) -> Union[Tuple[str, Any], None]:
        """returns [(name, val)...] or None"""
        if choices is None:
            return None
        return [_expand_choice(c, self.dtype) for c in choices]

    def __str__(self):
        return f"{self.name} = {self.val}"

    def __repr__(self):
        return f"LQ: {self.name} = {self.val}"

    def read_from_hardware(self, send_signal: bool = True):
        self.log.debug(f"{self.name}: read_from_hardware send_signal={send_signal}")
        if self.hardware_read_func is not None:
            with self.lock:
                self.oldval = self.val
                val = self.hardware_read_func()
            self.update_value(
                new_val=val, update_hardware=False, send_signal=send_signal
            )
        else:
            self.log.warning(
                f"{self.name} read_from_hardware called when not connected to hardware"
            )
        return self.val

    def write_to_hardware(self, reread_hardware=None):
        if reread_hardware is None:
            # if undefined, default to stored reread_from_hardware_after_write bool
            reread_hardware = self.reread_from_hardware_after_write
        # Read from Hardware
        if self.has_hardware_write():
            with self.lock:
                self.hardware_set_func(self.val)
            if reread_hardware:
                self.read_from_hardware(send_signal=False)

    @property
    def value(self):
        "return stored value"
        return self.val

    @QtCore.Slot(str)
    @QtCore.Slot(float)
    @QtCore.Slot(int)
    @QtCore.Slot(bool)
    @QtCore.Slot()
    def update_value(
        self, new_val=None, update_hardware=True, send_signal=True, reread_hardware=None
    ):
        """
        Update stored value with new_val

        Change value of LQ and emit signals to inform listeners of change

        if *update_hardware* is true: call connected hardware write function

        =============== =================================================================================================================
        **Arguments:**  **Description:**
        new_val         New value for the LoggedQuantity to store
        update_hardware calls hardware_set_func if defined (default True)
        send_signal     sends out QT signals on upon change (default True)
        reread_hardware read from hardware after writing to hardware to ensure change (defaults to self.reread_from_hardware_after_write)
        =============== =================================================================================================================

        :returns: None

        """
        # use a thread lock during update_value to avoid another thread
        # calling update_value during the update_value

        if reread_hardware is None:
            # if undefined, default to stored reread_from_hardware_after_write bool
            reread_hardware = self.reread_from_hardware_after_write

        with self.lock:

            # sometimes a the sender is a textbox that does not send its new value,
            # grab the text() from it instead
            if new_val is None:
                if hasattr(self.sender(), "text"):
                    new_val = self.sender().text()

            self.oldval = self.coerce_to_type(self.val)
            new_val = self.coerce_to_type(new_val)

            self.log.debug(
                f"{self.path}: update_value {repr(self.oldval)} --> {repr(new_val)}  from sender:{repr(self.sender())}"
            )

            # check for equality of new vs old, do not proceed if they are same
            if self.same_values(self.oldval, new_val):
                self.log.debug(f"{self.path}: same_value so returning")
                return
            # else:
            #     self.log.debug(f"{self.path}: different values {self.oldval} {new_val}")

            # actually change internal state value and store prev. values
            self.prev_vals.appendleft(self.val)
            self.val = new_val

        # Read from Hardware
        if update_hardware and self.hardware_set_func:
            self.hardware_set_func(self.val)
            if reread_hardware:
                self.read_from_hardware(send_signal=False)
        # Send Qt Signals
        if send_signal:
            self.send_display_updates()

    def send_display_updates(self, force=False):
        """
        Emit updated_value signals if value has changed.

        =============  =============================================
        **Arguments**  **Description**
        *force*        will emit signals regardless of value change.
        =============  =============================================

        :returns: None

        """
        # self.log.debug("{self.name}:send_display_updates: {force=}. From {self.oldval} to {self.val}")
        if (not self.same_values(self.oldval, self.val)) or (force):
            self.updated_value[()].emit()

            str_val = self.string_value()
            self.updated_value[str].emit(str_val)
            self.updated_text_value.emit(str_val)

            if self.dtype in [float, int]:
                self.updated_value[float].emit(self.val)
                self.updated_value[int].emit(int(self.val))
            self.updated_value[bool].emit(bool(self.val))

            if self.choices is not None:
                choice_vals = [c[1] for c in self.choices]
                if self.val in choice_vals:
                    self.updated_choice_index_value.emit(choice_vals.index(self.val))
            self.oldval = self.val
        else:
            # no updates sent
            pass

    def same_values(self, v1, v2):
        """
        Compares two values of the LQ type, used in update_value

        =============  ====================
        **Arguments**  **Description**
        v1             value 1
        v2             value 2
        =============  ====================

        :returns: Boolean value (True or False)

        """
        return v1 == v2

    def string_value(self):
        if self.dtype == str:
            return self.val
        else:
            return self.fmt % self.val

    def ini_string_value(self):
        """
        :returns: A string showing the logged quantity value.
        """
        return str(self.val)

    def update_choice_index_value(self, new_choice_index, **kwargs):
        self.update_value(self.choices[new_choice_index][1], **kwargs)

    def add_listener(self, func, argtype=(), **kwargs):
        """
        Connect 'func' as a listener (Qt Slot) for the
        updated_value signal.
        By default 'func' should take no arguments,
        but argtype can define the data type that it should accept.
        but should be limited to those supported by LoggedQuantity
        (i.e. int, float, str)
        **kwargs are passed to the connect function
        appends the 'func' to the 'listeners' list
        """

        #    --> This is now handled by redefining sys.excepthook handle in ScopeFoundry.base_app
        # Wraps func in a try block to absorb the Exception to avoid crashing PyQt5 >5.5
        # see https://riverbankcomputing.com/pipermail/pyqt/2016-March/037134.html
        #         def wrapped_func(func):
        #             def f(*args):
        #                 try:
        #                     func(*args)
        #                 except Exception as err:
        #                     print "Exception on listener:"

        self.updated_value[argtype].connect(func, **kwargs)
        self.listeners.append(func)

    def change_readonly_on(self, other_lq, func=bool):
        """
        When the other_lq updates, this LQ will be change its readonly status to func(other_lq.val).
        """
        other_lq.add_listener(lambda: self.change_readonly(func(other_lq.val)))

    def connect_bidir_to_widget(self, widget):
        # DEPRECATED
        return self.connect_to_widget(widget)

    def set_widget_toolTip(self, widget: QtWidgets.QWidget, text=None):
        if not hasattr(widget, "setToolTip"):
            return []
        tips = [f"<b>{self.path}</b>"]
        if self.unit:
            tips.append(f"<i>({self.unit})</i>")
        for s in [self.description, widget.toolTip(), text]:
            if s is not None:
                tips.append(str(s))
        if self.protected:
            tips.append("\n<b>🔒 protected</b>")

        tips_str = "\n".join(tips)
        widget.setToolTip(f"<div style='white-space: pre-line;'>{tips_str}</div>")
        return tips

    def connect_to_widget(self, widget: QtWidgets.QWidget):
        """
        Creates Qt signal-slot connections between LQ and the QtWidget *widget*

        connects updated_value signal to the appropriate slot depending on
        the type of widget

        Makes a bidirectional connection to a QT widget, ie when LQ is updated,
        widget gets a signal and when widget is updated, the LQ receives a
        signal and update_value() slot is called.

        Handles many types of widgets:
         * QDoubleSpinBox
         * QCheckBox
         * QLineEdit
         * QComboBox
         * pyqtgraph.widgets.SpinBox.SpinBox

        =============  ====================================================================
        **Arguments**  **Description**
        widget         The name of the Qt GUI Object, examples of which are listed above.
                       For example, if you have a QDoubleSpinBox in the gui which you
                       renamed int_value_doubleSpinBox in the Qt Designer Object Inspector,
                       use self.ui.int_value_doubleSpinBox
        =============  ====================================================================

        :returns: None

        """

        self.set_widget_toolTip(widget)
        widget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        widget.customContextMenuRequested.connect(self.on_right_click)

        if isinstance(widget, QtWidgets.QDoubleSpinBox):

            widget.setKeyboardTracking(False)
            if self.vmin is not None:
                widget.setMinimum(self.vmin)
            if self.vmax is not None:
                widget.setMaximum(self.vmax)
            if self.unit is not None:
                widget.setSuffix(f" {self.unit}")
            widget.setDecimals(self.spinbox_decimals)
            widget.setSingleStep(self.spinbox_step)
            widget.setValue(self.val)

            @QtCore.Slot(float)
            def update_widget_value(x=None):
                """
                block signals from widget when value is set via lq.update_value.
                This prevents signal-slot loops between widget and lq
                """
                try:
                    widget.blockSignals(True)
                    widget.setValue(self.val)
                finally:
                    widget.blockSignals(False)

            # self.updated_value[float].connect(widget.setValue)
            widget.to_connection = self.updated_value[float].connect(
                update_widget_value
            )
            # if not self.ro:
            widget.valueChanged[float].connect(self.update_value)

        elif isinstance(widget, MinMaxQSlider):
            self.updated_value[float].connect(widget.update_value)
            widget.updated_value.connect(self.update_value)
            if self.unit is not None:
                widget.unit = self.unit
            widget.set_range(self.vmin, self.vmax)
            widget.set_name(self.name)

        elif isinstance(widget, QtWidgets.QSlider):
            SLIDER_MIN = -2147483648
            SLIDER_MAX = 2147483647
            vmin = min(max(self.vmin, SLIDER_MIN), SLIDER_MAX)
            vmax = max(min(self.vmax, SLIDER_MAX), SLIDER_MIN)
            int_overflow = vmin != self.vmin or vmax != self.vmax

            if int_overflow:
                self.log.info(
                    f"WARNING in using QSlider with {self.path}: Consider narrowing the range with vmin, vmax"
                )

            if self.dtype == int and not int_overflow:
                widget.setRange(int(vmin), int(vmax))
                self.updated_value[int].connect(widget.setValue)
                widget.valueChanged[int].connect(self.update_value)
                widget.setValue(self.value)

            elif self.dtype == float or (self.dtype == int and int_overflow):
                # the sliders values are scaled version of the actual value
                # making slider values symmetric around 0
                SLIDER_MIN = -2147483647
                self.vrange = self.vmax - self.vmin
                widget.setRange(SLIDER_MIN, SLIDER_MAX)
                slider_range = SLIDER_MAX - SLIDER_MIN

                def transform_to_slider(x):
                    lq_avg = 0.5 * (self.vmax + self.vmin)
                    self.vrange = self.vmax - self.vmin
                    return int(slider_range / self.vrange * (x - lq_avg))

                def transform_from_slider(x):
                    lq_avg = 0.5 * (self.vmax + self.vmin)
                    self.vrange = self.vmax - self.vmin
                    return self.vrange / slider_range * x + lq_avg

                # for float testing in tests/unittest/lq_connect_to_widget_test
                self._transform_to_slider = transform_to_slider
                self._transform_from_slider = transform_from_slider

                @QtCore.Slot(float)
                def update_widget_value(x):
                    """
                    block signals from widget when value is set via lq.update_value.
                    This prevents signal-slot loops between widget and lq
                    """
                    try:
                        widget.blockSignals(True)
                        widget.setValue(transform_to_slider(self.val))
                    finally:
                        widget.blockSignals(False)

                @QtCore.Slot(int)
                def update_lq(x):
                    self.update_value(transform_from_slider(x))

                widget.setSingleStep(1)
                self.updated_value[float].connect(update_widget_value)
                widget.valueChanged[int].connect(update_lq)
                widget.setValue(transform_to_slider(self.val))

            else:
                self.log.warning(
                    f"QSlider not supported with setting {self.path} of dtype={self.dtype}"
                )

        elif isinstance(widget, QtWidgets.QCheckBox):

            @QtCore.Slot(bool)
            def update_widget_value(x=None):
                try:
                    widget.blockSignals(True)
                    widget.setChecked(self.val)
                finally:
                    widget.blockSignals(False)

            self.updated_value[bool].connect(update_widget_value)

            if self.ro:
                widget.setEnabled(False)
            else:
                widget.clicked[bool].connect(self.update_value)
                # only works if widget was clicked or
                # widget value changed and widget.clicked.emit() was called
                # Maybe more rigourous would be
                # widget.stateChanged.connect(self.update_value)
                # but then for widgets in tristate can send states 0, 1, 2.
                # Do not know when state occurs

            if self.colors is not None:
                s = f"""QCheckBox:!checked {{ background: {self.colors[0]} }}
                        QCheckBox:checked  {{ background: {self.colors[-1]} }}"""
                widget.setStyleSheet(widget.styleSheet() + s)

        elif isinstance(widget, QtWidgets.QLineEdit):
            self.updated_text_value[str].connect(widget.setText)
            self.updated_value[str].connect(widget.setText)

            if self.ro:
                widget.setReadOnly(True)  # FIXME

            def on_edit_finished():
                self.log.debug(f"{self.path} qLineEdit on_edit_finished")
                try:
                    widget.blockSignals(True)
                    self.update_value(widget.text())
                finally:
                    widget.blockSignals(False)

            widget.editingFinished.connect(on_edit_finished)

        elif isinstance(widget, QtWidgets.QPlainTextEdit):
            # TODO Read only

            def on_lq_changed(new_text):
                current_cursor = widget.textCursor()
                current_cursor_pos = current_cursor.position()
                # print('current_cursor', current_cursor, current_cursor.position())
                widget.document().setPlainText(new_text)
                current_cursor.setPosition(current_cursor_pos)
                widget.setTextCursor(current_cursor)
                # print('current_cursor', current_cursor, current_cursor.position())

            def on_widget_textChanged():
                try:
                    widget.blockSignals(True)
                    self.update_value(widget.toPlainText())
                finally:
                    widget.blockSignals(False)

            # self.updated_text_value[str].connect(widget.document().setPlainText)
            self.updated_text_value[str].connect(on_lq_changed)
            widget.textChanged.connect(on_widget_textChanged)

        elif isinstance(widget, QtWidgets.QTextEdit):

            def on_updated(x=None):
                widget.setText(self.val)

            self.updated_value[str].connect(on_updated)

            def on_text_changed(x=None):
                try:
                    widget.blockSignals(True)
                    self.update_value(widget.toPlainText())
                finally:
                    widget.blockSignals(False)

            widget.textChanged.connect(on_text_changed)

        elif isinstance(widget, QtWidgets.QComboBox):
            assert self.choices is not None
            widget.clear()  # removes all old choices

            for choice_name, choice_value in self.choices:
                widget.addItem(choice_name, choice_value)
            self.updated_choice_index_value[int].connect(widget.setCurrentIndex)
            widget.currentIndexChanged.connect(self.update_choice_index_value)

            if self.colors is not None:
                for i, qcolor in enumerate(self.qcolors):
                    widget.setItemData(i, qcolor, QtCore.Qt.BackgroundRole)

                def update_background_color(idx):
                    if idx < len(self.qcolors):
                        qc = self.qcolors[idx]
                    else:
                        qc = QtGui.QColor("lightgrey")

                    s = f"""QComboBox{{
                                selection-background-color: {qc.name()};
                                selection-color: black;
                                background: {qc.name()};
                            }}"""
                    widget.setStyleSheet(widget.styleSheet() + s)

                widget.currentIndexChanged.connect(update_background_color)

        elif isinstance(widget, pg.widgets.SpinBox.SpinBox):
            # widget.setFocusPolicy(QtCore.Qt.StrongFocus)
            suffix = self.unit
            if self.unit is None:
                suffix = ""
            if self.dtype == int:
                integer = True
                minStep = 1
                step = 1
            else:
                integer = False
                minStep = 0.1
                step = 0.1
            opts = dict(
                suffix=suffix,
                siPrefix=True,
                dec=True,
                step=step,
                minStep=minStep,
                bounds=[self.vmin, self.vmax],
                int=integer,
            )
            if self.si:
                del opts["step"]
                del opts["minStep"]

            widget.setOpts(**opts)

            if self.ro:
                widget.setEnabled(False)
                widget.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
                widget.setReadOnly(True)
            # widget.setDecimals(self.spinbox_decimals)
            if not self.si:
                widget.setSingleStep(self.spinbox_step)

            # self.updated_value[float].connect(widget.setValue)
            # if not self.ro:
            # widget.valueChanged[float].connect(self.update_value)

            @QtCore.Slot(float)
            def update_widget_value(x):
                """
                block signals from widget when value is set via lq.update_value.
                This prevents signal loops
                """
                try:
                    widget.blockSignals(True)
                    widget.setValue(x)
                finally:
                    widget.blockSignals(False)

            self.updated_value[float].connect(update_widget_value)

            def on_widget_update(_widget):
                self.update_value(_widget.value())

            widget.sigValueChanged.connect(on_widget_update)

        elif isinstance(widget, QtWidgets.QLabel):
            self.updated_text_value.connect(widget.setText)
        elif isinstance(widget, QtWidgets.QProgressBar):

            def set_progressbar(x, widget=widget):
                self.log.debug(f"set_progressbar {x}")
                widget.setValue(int(x))

            self.updated_value.connect(set_progressbar)
        elif isinstance(widget, QtWidgets.QLCDNumber):
            self.updated_value[(self.dtype)].connect(widget.display)
        else:
            raise ValueError("Unknown widget type")

        self.send_display_updates(force=True)
        # self.widget = widget
        self.widget_list.append(widget)
        self.change_readonly(self.ro)

    def disconnect_from_widget(self, widget: QtWidgets.QWidget):
        """
        Disconnects the Qt signal-slot connections between LQ and the QtWidget *widget*

        :returns: None

        """
        if widget not in self.widget_list:
            # print(f"Widget {widget} not connected to {self.name}")
            return

        self.widget_list.remove(widget)

        widget.disconnect()

        if hasattr(widget, "to_connection"):
            if isinstance(widget, QtWidgets.QDoubleSpinBox):
                self.updated_value[float].disconnect(widget.to_connection)
            # TODO: other widgets!

            # elif isinstance(widget, QtWidgets.QCheckBox):
            #     self.updated_value[bool].disconnect(widget.setChecked)
            # elif isinstance(widget, QtWidgets.QLineEdit):
            #     self.updated_value[str].disconnect(widget.setText)
            #     self.updated_text_value[str].disconnect(widget.setText)
            # elif isinstance(widget, QtWidgets.QPlainTextEdit):
            #     self.updated_text_value[str].disconnect(widget.document().setPlainText)
            #     self.updated_value[str].disconnect(widget.document().setPlainText)
            # elif isinstance(widget, QtWidgets.QSlider):
            #     self.updated_value[float].disconnect(widget.setValue)
            # elif isinstance(widget, MinMaxQSlider):
            #     self.updated_value[float].disconnect(widget.update_value)
            # elif isinstance(widget, QtWidgets.QLCDNumber):
            #     self.update_value[float].disconnect(widget.display)
            # elif isinstance(widget, QtWidgets.QProgressBar):
            #     self.updated_value.disconnect(widget.setValue)
            # elif isinstance(widget, QtWidgets.QComboBox):
            #     self.updated_choice_index_value[int].disconnect(widget.setCurrentIndex)
            #     widget.currentIndexChanged.disconnect(self.update_choice_index_value)

    def connect_to_widget_one_way(self, widget):
        """
        Creates Qt signal-slot connections between LQ and the QtWidget *widget* ONLY IN ONE WAY. CHANGES IN WIDGET WILL NOT TRANSFER TO LQ!!!
        rest is the same as connect_to_widget

        """

        self.set_widget_toolTip(widget)

        if isinstance(widget, QtWidgets.QDoubleSpinBox):

            widget.setKeyboardTracking(False)
            if self.vmin is not None:
                widget.setMinimum(self.vmin)
            if self.vmax is not None:
                widget.setMaximum(self.vmax)
            if self.unit is not None:
                widget.setSuffix(f" {self.unit}")
            widget.setDecimals(self.spinbox_decimals)
            widget.setSingleStep(self.spinbox_step)
            widget.setValue(self.val)

            # events
            @QtCore.Slot(float)
            def update_widget_value(x):
                """
                block signals from widget when value is set via lq.update_value.
                This prevents signal-slot loops between widget and lq
                """
                try:
                    widget.blockSignals(True)
                    widget.setValue(self.val)
                finally:
                    widget.blockSignals(False)

            # self.updated_value[float].connect(widget.setValue)
            self.updated_value[float].connect(update_widget_value)

        elif isinstance(widget, MinMaxQSlider):
            self.updated_value[float].connect(widget.update_value)
            if self.unit is not None:
                widget.unit = self.unit
            widget.set_range(self.vmin, self.vmax)
            widget.set_name(self.name)

        elif isinstance(widget, QtWidgets.QSlider):

            SLIDER_MIN = -2147483648
            SLIDER_MAX = 2147483647
            vmin = min(max(self.vmin, SLIDER_MIN), SLIDER_MAX)
            vmax = max(min(self.vmax, SLIDER_MAX), SLIDER_MIN)
            int_overflow = vmin != self.vmin or vmax != self.vmax

            if int_overflow:
                self.log.info(
                    f"WARNING in using QSlider with {self.path}: Consider narrowing the range with vmin, vmax"
                )

            if self.dtype == int and not int_overflow:
                widget.setRange(int(vmin), int(vmax))
                self.updated_value[int].connect(widget.setValue)
                widget.setValue(self.value)

            elif self.dtype == float or (self.dtype == int and int_overflow):
                # the sliders values are scaled version of the actual value
                # making slider values symmetric around 0
                SLIDER_MIN = -2147483647
                self.vrange = self.vmax - self.vmin
                widget.setRange(SLIDER_MIN, SLIDER_MAX)
                slider_range = SLIDER_MAX - SLIDER_MIN

                def transform_to_slider(x):
                    lq_avg = 0.5 * (self.vmax + self.vmin)
                    self.vrange = self.vmax - self.vmin
                    return int(slider_range / self.vrange * (x - lq_avg))

                def transform_from_slider(x):
                    lq_avg = 0.5 * (self.vmax + self.vmin)
                    self.vrange = self.vmax - self.vmin
                    return self.vrange / slider_range * x + lq_avg

                # for float testing in tests/unittest/lq_connect_to_widget_test
                self._transform_to_slider = transform_to_slider
                self._transform_from_slider = transform_from_slider

                @QtCore.Slot(float)
                def update_widget_value(x):
                    """
                    block signals from widget when value is set via lq.update_value.
                    This prevents signal-slot loops between widget and lq
                    """
                    try:
                        widget.blockSignals(True)
                        widget.setValue(transform_to_slider(self.val))
                    finally:
                        widget.blockSignals(False)

                widget.setSingleStep(1)
                self.updated_value[float].connect(update_widget_value)
                widget.setValue(transform_to_slider(self.val))

            else:
                self.log.warning(
                    f"QSlider not supported with setting {self.path} of dtype={self.dtype}"
                )

        elif isinstance(widget, QtWidgets.QCheckBox):

            @QtCore.Slot(bool)
            def update_widget_value(x=None):
                try:
                    widget.blockSignals(True)
                    widget.setChecked(self.val)
                finally:
                    widget.blockSignals(False)

            self.updated_value[bool].connect(update_widget_value)

            if self.ro:
                widget.setEnabled(False)

            if self.colors is not None:
                s = f"""QCheckBox:!checked {{ background: {self.colors[0]} }}
                        QCheckBox:checked  {{ background: {self.colors[-1]} }}"""
                widget.setStyleSheet(widget.styleSheet() + s)

        elif isinstance(widget, QtWidgets.QLineEdit):
            self.updated_text_value[str].connect(widget.setText)
            self.updated_value[str].connect(widget.setText)
            if self.ro:
                widget.setReadOnly(True)  # FIXME

        elif isinstance(widget, QtWidgets.QPlainTextEdit):
            # TODO Read only

            def on_lq_changed(new_text):
                current_cursor = widget.textCursor()
                current_cursor_pos = current_cursor.position()
                # print('current_cursor', current_cursor, current_cursor.position())
                widget.document().setPlainText(new_text)
                current_cursor.setPosition(current_cursor_pos)
                widget.setTextCursor(current_cursor)
                # print('current_cursor', current_cursor, current_cursor.position())

            # self.updated_text_value[str].connect(widget.document().setPlainText)
            self.updated_text_value[str].connect(on_lq_changed)

        elif isinstance(widget, QtWidgets.QComboBox):
            # need to have a choice list to connect to a QComboBox
            assert self.choices is not None
            widget.clear()  # removes all old choices
            for choice_name, choice_value in self.choices:
                widget.addItem(choice_name, choice_value)
            self.updated_choice_index_value[int].connect(widget.setCurrentIndex)

            if self.colors is not None:
                for i, qcolor in enumerate(self.qcolors):
                    widget.setItemData(i, qcolor, QtCore.Qt.BackgroundRole)

                def update_background_color(idx):
                    if idx < len(self.qcolors):
                        qc = self.qcolors[idx]
                    else:
                        qc = QtGui.QColor("white")

                    s = f"""QComboBox{{
                                selection-background-color: {qc.name()};
                                selection-color: black;
                                background: {qc.name()};
                                }}"""
                    widget.setStyleSheet(widget.styleSheet() + s)

                widget.currentIndexChanged.connect(update_background_color)

        elif isinstance(widget, pg.widgets.SpinBox.SpinBox):
            # widget.setFocusPolicy(QtCore.Qt.StrongFocus)
            suffix = self.unit
            if self.unit is None:
                suffix = ""
            if self.dtype == int:
                integer = True
                minStep = 1
                step = 1
            else:
                integer = False
                minStep = 0.1
                step = 0.1
            opts = dict(
                suffix=suffix,
                siPrefix=True,
                dec=True,
                step=step,
                minStep=minStep,
                bounds=[self.vmin, self.vmax],
                int=integer,
            )
            if self.si:
                del opts["step"]
                del opts["minStep"]

            widget.setOpts(**opts)

            if self.ro:
                widget.setEnabled(False)
                widget.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
                widget.setReadOnly(True)
            # widget.setDecimals(self.spinbox_decimals)
            if not self.si:
                widget.setSingleStep(self.spinbox_step)

            # self.updated_value[float].connect(widget.setValue)
            # if not self.ro:
            # widget.valueChanged[float].connect(self.update_value)

            @QtCore.Slot(float)
            def update_widget_value(x):
                """
                block signals from widget when value is set via lq.update_value.
                This prevents signal loops
                """
                try:
                    widget.blockSignals(True)
                    widget.setValue(x)
                finally:
                    widget.blockSignals(False)

            self.updated_value[float].connect(update_widget_value)

        elif isinstance(widget, QtWidgets.QLabel):
            self.updated_text_value.connect(widget.setText)

        elif isinstance(widget, QtWidgets.QProgressBar):

            @QtCore.Slot(int)
            @QtCore.Slot(float)
            def set_progressbar(x, widget=widget):
                self.log.debug(f"set_progressbar {x}")
                widget.setValue(int(x))

            self.updated_value.connect(set_progressbar)

        elif isinstance(widget, QtWidgets.QLCDNumber):
            self.updated_value[(self.dtype)].connect(widget.display)
        else:
            raise ValueError("Unknown widget type")

        self.send_display_updates(force=True)
        # self.widget = widget
        self.widget_list.append(widget)
        self.change_readonly(self.ro)

    def connect_to_pushButton(
        self,
        pushButton: QtWidgets.QPushButton,
        colors=("rgba(105, 240, 104, 220)", "rgba(255, 49, 47, 220)"),
        texts=("▶ START", "🛑 STOP"),
        styleSheet_amendment="""
        QPushButton{ padding:3px }
        QPushButton:hover:!pressed{ border: 1px solid black; }
        """,
    ):
        assert isinstance(pushButton, QtWidgets.QPushButton)
        assert self.dtype == bool

        pushButton.setCheckable(True)

        @QtCore.Slot(bool)
        def update_widget_value(x=None):
            try:
                pushButton.blockSignals(True)
                pushButton.setChecked(self.val)
                pushButton.setText(texts[int(x)])
            finally:
                pushButton.blockSignals(False)

        self.updated_value[bool].connect(update_widget_value)
        pushButton.toggled[bool].connect(self.update_value)

        if self.colors and not colors:
            colors = self.colors
        if colors:
            s = f"""QPushButton:!checked{{ background:{colors[0]}; border: 1px solid grey; }}
                    QPushButton:checked{{ background:{colors[1]}; border: 1px solid grey; }}"""
            pushButton.setStyleSheet(pushButton.styleSheet() + s + styleSheet_amendment)

        if self.ro:
            pushButton.setEnabled(False)

        self.set_widget_toolTip(pushButton)
        self.send_display_updates(force=True)
        self.widget_list.append(pushButton)

    def connect_to_lq(self, lq):
        # FIXME, does not seem to work for dtype=int
        self.updated_value[(self.dtype)].connect(lq.update_value)
        lq.updated_value[(lq.dtype)].connect(self.update_value)

    def change_choice_list(self, choices, new_val=None):
        """
        =============  =================================================
        **Arguments**  **Description**
        *choices*      list of new choices
                       Either of the form [val_1, val_2, ...]
                       or [(name_1, val_1), ...]
        *new_val*      the value that is set after choices have been
                       changed. If None, the previous value is retained,
                       otherwise the last element of *choices*.
        =============  =================================================
        """

        with self.lock:
            self.choices = self._expand_choices(choices)
            for widget in self.widget_list:
                if not isinstance(widget, QtWidgets.QComboBox):
                    continue
                widget.blockSignals(True)
                widget.clear()
                if self.choices is not None:
                    for choice_name, choice_value in self.choices:
                        widget.addItem(choice_name, choice_value)
                widget.blockSignals(False)

        if not self.choices:
            return

        # set a value
        if new_val is None:
            new_val = self.val
        values = [x[1] for x in self.choices]
        if not new_val in values:
            new_val = values[0]
        self.update_value(new_val)

        self.send_display_updates(force=True)

    def add_choices(self, choices, allow_duplicates=False, new_val=None):
        if not isinstance(choices, (list, tuple)):
            choices = [choices]

        if not choices:
            return False

        new_choices = self.choices + self._expand_choices(choices)
        if not allow_duplicates:
            new_choices = remove_duplicates(new_choices)

        if new_val is None:
            new_val = self.val

        self.change_choice_list(new_choices, new_val)
        return True

    def remove_choices(self, choices, new_val=None):
        """
        =============  =================================================
        **Arguments**  **Description**
        *choices*      list of choices to be removed.
                       Either of the form [val_1, val_2, ...]
                       or [(name_1, val_1), ...]

        *new_val*      the value that is set after choices have been
                       changed. If None, the previous value is retained,
                       otherwise the last element of *choices*.
        =============  =================================================
        """
        if not isinstance(choices, (list, tuple)):
            choices = [choices]
        choices = [c for c in self.choices if c not in self._expand_choices(choices)]
        self.change_choice_list(choices, new_val)
        return True

    def change_min_max(self, vmin=-1e12, vmax=+1e12):
        # TODO  setRange should be a slot for the updated_min_max signal
        with self.lock:
            self.vmin = vmin
            self.vmax = vmax
            for widget in self.widget_list:  # may not work for certain widget types
                widget.setRange(vmin, vmax)
            self.updated_min_max.emit(vmin, vmax)

    def change_readonly(self, ro=True):
        with self.lock:
            self.ro = ro
            for widget in self.widget_list:
                if hasattr(widget, "setReadOnly"):
                    widget.setReadOnly(self.ro)
                else:
                    widget.setEnabled(not self.ro)
                # TODO other widget types
            self.updated_readonly.emit(self.ro)

    def change_unit(self, unit):
        with self.lock:
            self.unit = unit
            for widget in self.widget_list:
                if isinstance(widget, QtWidgets.QDoubleSpinBox):
                    if self.unit is not None:
                        widget.setSuffix(f" {self.unit}")

                elif isinstance(widget, pg.widgets.SpinBox.SpinBox):
                    # widget.setFocusPolicy(QtCore.Qt.StrongFocus)
                    suffix = self.unit
                    if self.unit is None:
                        suffix = ""
                    opts = dict(suffix=suffix)

                    widget.setOpts(**opts)

    def is_connected_to_hardware(self):
        """
        :returns: True if either self.hardware_read_func or
        self.hardware_set_func are defined. False if None.
        """
        return (self.hardware_read_func is not None) or (
            self.hardware_set_func is not None
        )

    def has_hardware_read(self):
        return self.hardware_read_func is not None

    def has_hardware_write(self):
        return self.hardware_set_func is not None

    def connect_to_hardware(self, read_func=None, write_func=None):
        if read_func is not None:
            assert callable(read_func)
            self.hardware_read_func = read_func
        if write_func is not None:
            assert callable(write_func)
            self.hardware_set_func = write_func

    def disconnect_from_hardware(self, dis_read=True, dis_write=True):
        if dis_read:
            self.hardware_read_func = None
        if dis_write:
            self.hardware_set_func = None

    def connect_lq_math(self, lqs, func, reverse_func=None):
        """
        Links LQ to other LQs using math functions

        takes a func that takes a set of logged quantities
        new_val = f(lq1, lq2, ...)

        when any of the lqs change, the value of this derived LQ
        will be updated based on func

        reverse_func allows changes to this LQ to update lqs via reverse func

        new_lq1_val, new_lq2_val, ... = g(new_val, old_lqs_values)
        or
        new_lq1_val, new_lq2_val, ... = g(new_val)

        """

        try:
            self.math_lqs = tuple(lqs)
        except TypeError:  # if not iterable, assume its a single LQ
            self.math_lqs = (lqs,)
        self.math_func = func

        self.reverse_math_func = reverse_func
        if reverse_func is not None:
            rev_sig = signature(reverse_func)
            self.reverse_func_num_params = len(rev_sig.parameters)

        def update_math():
            lq_vals = [lq.value for lq in self.math_lqs]
            new_val = self.math_func(*lq_vals)
            # print(self.name, "update_math", lq_vals, "-->", new_val, )
            self.update_value(new_val)

        if reverse_func:

            def update_math_reverse():
                lq_vals = [lq.value for lq in self.math_lqs]
                if self.reverse_func_num_params > 1:
                    new_vals = self.reverse_math_func(self.val, lq_vals)
                else:
                    new_vals = self.reverse_math_func(self.val)

                try:
                    new_vals = tuple(new_vals)
                except TypeError:  # if not iterable, assume its a single value
                    new_vals = (new_vals,)

                for lq, new_val in zip(self.math_lqs, new_vals):
                    lq.update_value(new_val)

        for lq in self.math_lqs:
            lq.updated_value[()].connect(update_math)
            if reverse_func:
                self.add_listener(update_math_reverse)

        update_math()

    def read_from_lq_math(self):
        lq_vals = [lq.value for lq in self.math_lqs]
        new_val = self.math_func(*lq_vals)
        # print("read_from_lq_math", lq_vals, "-->", new_val, )
        self.update_value(new_val)

    def connect_lq_scale(self, lq, scale):
        self.lq_scale = scale
        self.connect_lq_math(
            (lq,),
            func=lambda x: scale * x,
            reverse_func=lambda y, old_vals: [
                y * 1.0 / scale,
            ],
        )

    def new_default_widget(self):
        """returns the appropriate QWidget for the datatype of the
        LQ. automatically connects widget
        """
        if self.default_widget_factory is not None:
            widget = self.default_widget_factory()
        elif self.choices is not None:
            widget = QtWidgets.QComboBox()
        elif self.dtype in [int, float]:
            if self.si:
                widget = pg.SpinBox()
            else:
                widget = QtWidgets.QDoubleSpinBox()
        elif self.dtype in [bool]:
            widget = QtWidgets.QCheckBox()
        elif self.dtype in [str]:
            widget = QtWidgets.QLineEdit()
        self.connect_to_widget(widget)

        if not any((self.is_clipboardable, self.is_cmd)):
            return widget

        wrap_widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(wrap_widget)
        layout.addWidget(widget)
        layout.setSpacing(0)

        if self.is_clipboardable:
            copy_button = QtWidgets.QPushButton("⎘")
            copy_button.setMaximumWidth(25)
            copy_button.setToolTip("copy to clipboard")
            copy_button.clicked.connect(self.copy_to_clipboard)
            layout.addWidget(copy_button)

        if self.is_cmd:
            run_button = QtWidgets.QPushButton("➲")
            run_button.setMaximumWidth(33)
            run_button.setToolTip("run detached")
            run_button.clicked.connect(self.run_subprocess_detached)
            layout.addWidget(run_button)

        return wrap_widget

    def new_pushButton(
        self,
        texts=("▶ START", "🛑 STOP"),
        colors=("rgba(105, 240, 104, 220)", "rgba(255, 49, 47, 220)"),
        **kwargs,
    ):
        """kwargs is past to self.connect_to_pushButton():"""
        assert self.dtype == bool
        pushButton = QtWidgets.QPushButton()
        self.connect_to_pushButton(pushButton, colors, texts, **kwargs)
        return pushButton

    def new_pg_parameter(self):
        from pyqtgraph.parametertree import Parameter

        dtype_map = {str: "str", float: "float", int: "int", bool: "bool"}

        if self.choices:
            print(self.name, "have choices")
            print(self.choices)
            # choices should be tuple [ ('name', val) ... ] or simple list [val, val, ...]

            p = Parameter.create(
                name=self.name, type="list", values=dict(self.choices), value=self.value
            )

            def update_param(v):
                print("updating parameter", self.name, p, v)
                p.setValue(v)

            self.updated_value[self.dtype].connect(
                update_param
            )  # (lambda v, p=p: p.setValue(v))
            p.sigValueChanged.connect(lambda p, v: self.update_value(v))

            return p
        if self.is_array:
            # DOES NOT WORK CORRECTLY
            p = Parameter.create(name=self.name, type="str", value=str(self.value))

            return p
        else:
            p = Parameter.create(
                name=self.name, type=dtype_map[self.dtype], value=self.value
            )

            def update_param(v):
                print("updating parameter", self.name, p, v)
                p.setValue(v)

            self.updated_value[self.dtype].connect(
                update_param
            )  # (lambda v, p=p: p.setValue(v))
            p.sigValueChanged.connect(lambda p, v: self.update_value(v))

            return p

    def set_path(self, path: str):
        self.path = path

    def propose_value(self, name: str, value):
        self.proposed_values.appendleft((name, value))

    def on_right_click(self, position=None):
        cmenu = QtWidgets.QMenu()
        cmenu.addAction(f"{self.path}")
        cmenu.addSeparator()

        # prev. values
        cmenu.addAction("prev. values").setEnabled(False)
        for val in self.prev_vals:
            if self.ro:
                cmenu.addAction(f"{val}")
            else:
                cmenu.addAction(f"{val}", partial(self.update_value, new_val=val))

        # proposed values
        for proposed_name, val in self.proposed_values:
            cmenu.addSeparator()
            cmenu.addAction(proposed_name).setEnabled(False)
            if self.ro:
                cmenu.addAction(f"{val}")
            else:
                cmenu.addAction(f"{val}", partial(self.update_value, new_val=val))

        cmenu.exec_(QtGui.QCursor.pos())

    def connect_to_clipboardable(self, io_widget, copy_button: QtWidgets.QPushButton):
        """type of io_widget"""
        self.connect_to_widget(io_widget)
        copy_button.clicked.connect(self.copy_to_clipboard)

    def copy_to_clipboard(self):
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(self.val)

    def connect_to_run_subprocess_detached(
        self, line_edit: QtWidgets.QWidget, run_button: QtWidgets.QPushButton
    ):
        self.connect_to_widget(line_edit)
        run_button.clicked.connect(self.run_subprocess_detached)

    def run_subprocess_detached(self):
        subprocess.Popen(
            self.val.split(),
            shell=True,
            # stdin=subprocess.PIPE,
            # stdout=subprocess.PIPE,
            # stderr=subprocess.PIPE,
        )


def to_q_color(color, default="lightgrey"):
    try:
        qcolor = QtGui.QColor(color)
    except TypeError:
        qcolor = QtGui.QColor(default)
    finally:
        if qcolor.isValid():
            return qcolor
        return QtGui.QColor(default)


def _expand_choice(c, dtype):
    if isinstance(c, tuple):
        name, val = c
        return (str(name), dtype(val))
    elif isinstance(c, Enum):
        return (c.name, dtype(c.value))
    return (str(c), dtype(c))


# https://stackoverflow.com/questions/480214/how-do-i-remove-duplicates-from-a-list-while-preserving-order
def remove_duplicates(l: List) -> List:
    seen = set()
    seen_add = seen.add
    return [x for x in l if not (x in seen or seen_add(x))]



import logging
import time
import unittest

from ScopeFoundry import BaseMicroscopeApp, HardwareComponent


class FailHW(HardwareComponent):

    name = 'fail_hw'

    def setup(self):
        print(self.name, "setup")

        self.settings.New('fail_on_connect', dtype=bool)
        self.settings.New('fail_on_disconnect', dtype=bool, initial=False)

    def connect(self):
        print(self.name, "connect")  
        if self.settings['fail_on_connect']:     
            raise IOError("CONNECT FAIL!")

    def disconnect(self):
        print(self.name, "disconnect")        
        if self.settings["fail_on_disconnect"]:
            raise IOError("DISCONNECT FAIL!")


class App(BaseMicroscopeApp):

    name = 'test_app'

    def setup(self):

        self.add_hardware(FailHW(self))
        logging.getLogger("LoggedQuantity").setLevel('DEBUG')


class AppTest(unittest.TestCase):

    def setUp(self):
        # import sys

        self.app = App([])
        self.hw = self.app.hardware["fail_hw"]
        logging.getLogger("LoggedQuantity").setLevel("ERROR")

    def tearDown(self) -> None:
        self.app.on_close()
        del self.app
        return super().tearDown()

    def test_connect(self):
        self.hw.settings["fail_on_connect"] = False
        self.hw.settings["fail_on_disconnect"] = False
        # self.hw.settings["connected"] = True
        self.hw.enable_connection(True)
        self.assertTrue(self.hw.settings["connected"])

    def test_connect_failure(self):
        self.hw.settings["fail_on_connect"] = True
        self.hw.settings["fail_on_disconnect"] = False
        with self.assertRaises(IOError):
            self.hw.enable_connection(True)
        time.sleep(0.1)
        self.assertFalse(self.hw.settings["connected"])

    def test_disconnect_failure(self):
        self.hw.settings["fail_on_connect"] = False
        self.hw.settings["fail_on_disconnect"] = True
        with self.assertRaises(IOError):
            self.hw.enable_connection(True)
            self.hw.enable_connection(False)
        # self.assertTrue(self.hw.settings["connected"]) # Behavior as of now. But what behaviour do we actully want here?


if __name__ == '__main__':
    # unittest.main()
    import sys
    app = App(sys.argv)

    app.exec_()



import time

import numpy as np

from ScopeFoundry import h5_io

from .base_raster_scan import BaseRaster2DScan


class BaseRaster2DFrameSlowScan(BaseRaster2DScan):

    name = "base_raster_2D_frame_slowscan"

    def run(self):
        S = self.settings

        # Hardware
        # self.apd_counter_hc = self.app.hardware_components['apd_counter']
        # self.apd_count_rate = self.apd_counter_hc.apd_count_rate
        # self.stage = self.app.hardware_components['dummy_xy_stage']

        # Data File
        # H5

        # Compute data arrays
        self.compute_scan_arrays()

        self.initial_scan_setup_plotting = True

        self.display_image_map = np.zeros(self.scan_shape, dtype=float)
        self.pixel_times = np.zeros(self.scan_shape, dtype=float)

        # h5 data file setup
        self.t0 = time.time()

        if self.settings["save_h5"]:
            self.h5_file = h5_io.h5_base_file(self.app, measurement=self)

            self.h5_file.attrs["time_id"] = self.t0
            H = self.h5_meas_group = h5_io.h5_create_measurement_group(
                self, self.h5_file
            )

            # create h5 data arrays
            H["h_array"] = self.h_array
            H["v_array"] = self.v_array
            H["range_extent"] = self.range_extent
            H["corners"] = self.corners
            H["imshow_extent"] = self.imshow_extent
            H["scan_h_positions"] = self.scan_h_positions
            H["scan_v_positions"] = self.scan_v_positions
            H["scan_slow_move"] = self.scan_slow_move
            H["scan_index_array"] = self.scan_index_array
            self.pixel_times_h5 = self.create_h5_framed_dataset(
                name="pixel_times", single_frame_map=self.pixel_times, dtype=float
            )

        self.frame_i = 0
        self.pixel_i = 0
        self.current_scan_index = self.scan_index_array[0]

        self.pre_scan_setup()

        try:
            while not self.interrupt_measurement_called:
                # start scan
                for i in range(self.settings["n_frames"]):
                    if self.settings["save_h5"]:
                        self.extend_h5_framed_dataset(self.pixel_times_h5, self.frame_i)
                    if self.interrupt_measurement_called:
                        break
                    # start frame
                    self.pixel_i = 0
                    self.current_scan_index = self.scan_index_array[0]
                    self.move_position_start(
                        self.scan_h_positions[0], self.scan_v_positions[0]
                    )
                    self.on_new_frame(self.frame_i)

                    for self.pixel_i in range(self.Npixels):
                        if self.interrupt_measurement_called:
                            break

                        i = self.pixel_i

                        self.current_scan_index = self.scan_index_array[i]
                        kk, jj, ii = self.current_scan_index

                        h, v = self.scan_h_positions[i], self.scan_v_positions[i]

                        if self.pixel_i == 0:
                            dh = 0
                            dv = 0
                        else:
                            dh = self.scan_h_positions[i] - self.scan_h_positions[i - 1]
                            dv = self.scan_v_positions[i] - self.scan_v_positions[i - 1]

                        if self.scan_slow_move[i]:
                            self.move_position_slow(h, v, dh, dv)
                            if self.settings["save_h5"]:
                                self.h5_file.flush()  # flush data to file every slow move
                            # self.app.qtapp.ProcessEvents()
                            time.sleep(0.01)
                        else:
                            self.move_position_fast(h, v, dh, dv)

                        self.pos = (h, v)
                        # each pixel:
                        # acquire signal and save to data array
                        pixel_t0 = time.time()
                        self.pixel_times[kk, jj, ii] = pixel_t0
                        if self.settings["save_h5"]:
                            self.pixel_times_h5[self.frame_i, kk, jj, ii] = pixel_t0
                        self.collect_pixel(self.pixel_i, self.frame_i, kk, jj, ii)
                        S["progress"] = (
                            100.0
                            * (self.frame_i * self.Npixels + self.pixel_i)
                            / (self.Npixels * self.settings["n_frames"])
                        )
                    self.on_end_frame(self.frame_i)
                    self.frame_i += 1
                if not self.settings["continuous_scan"]:
                    break
        finally:
            self.post_scan_cleanup()
            if self.settings["save_h5"] and hasattr(self, "h5_file"):
                self.h5_file.close()

    def move_position_start(self, x, y):
        self.stage.settings["x_position"] = x
        self.stage.settings["y_position"] = y

    def move_position_slow(self, x, y, dx, dy):
        self.stage.settings["x_position"] = x
        self.stage.settings["y_position"] = y

    def move_position_fast(self, x, y, dx, dy):
        self.stage.settings["x_position"] = x
        self.stage.settings["y_position"] = y
        # x = self.stage.settings['x_position']
        # y = self.stage.settings['y_position']
        # x = self.stage.settings.get_lq("x_position").read_from_hardware()
        # y = self.stage.settings.get_lq("y_position").read_from_hardware()
        # print(x,y)

    def pre_scan_setup(self):
        print(self.name, "pre_scan_setup not implemented")
        # hardware
        # create data arrays
        # update figure

    def collect_pixel(self, pixel_num, frame_i, k, j, i):
        # collect data
        # store in arrays
        print(
            self.name, "collect_pixel", pixel_num, frame_i, k, j, i, "not implemented"
        )

    def post_scan_cleanup(self):
        print(self.name, "post_scan_setup not implemented")

    def on_new_frame(self, frame_i):
        pass

    def on_end_frame(self, frame_i):
        pass

    @property
    def frames_scan_shape(self):
        """
        Returns the shape of data arrays for n_frames:
        (n_frames, N_subframes, Nv, Nh)
        """
        return (self.settings["n_frames"],) + self.scan_shape

    def create_h5_framed_dataset(self, name, single_frame_map, **kwargs):
        """
        Create and return an empty HDF5 dataset in self.h5_meas_group that can store
        multiple frames of single_frame_map.

        Must fill the dataset as frames roll in.

        creates reasonable defaults for compression and dtype, can be overriden
        with**kwargs are sent directly to create_dataset
        """
        if self.settings["save_h5"]:
            shape = (self.settings["n_frames"],) + single_frame_map.shape
            if self.settings["continuous_scan"]:
                # allow for array to grow to store additional frames
                maxshape = (None,) + single_frame_map.shape
            else:
                maxshape = shape
            # print('maxshape', maxshape)
            default_kwargs = dict(
                name=name,
                shape=shape,
                dtype=single_frame_map.dtype,
                # chunks=(1,),
                maxshape=maxshape,
                compression="gzip",
                # shuffle=True,
            )
            default_kwargs.update(kwargs)
            map_h5 = self.h5_meas_group.create_dataset(**default_kwargs)
            return map_h5

    def extend_h5_framed_dataset(self, map_h5, frame_num):
        """
        Adds additional frames to dataset map_h5, if frame_num
        is too large. Adds n_frames worth of extra frames
        """
        if self.settings["continuous_scan"]:
            current_num_frames = map_h5.shape[0]
            frame_shape = map_h5.shape[1:]
            if frame_num >= current_num_frames:
                # print ("extend_h5_framed_dataset", map_h5.name, map_h5.shape, frame_num)
                n_frames_extend = self.settings["n_frames"]
                new_num_frames = n_frames_extend * (1 + frame_num // n_frames_extend)
                map_h5.resize((new_num_frames,) + tuple(frame_shape))
                return True
            else:
                # "Dataset is large enough, no expansion"
                return False
        else:
            # "non continuous scan, no expansion"
            return False



import unittest
from unittest.mock import Mock

from qtpy.QtWidgets import QApplication

from ScopeFoundry.operations import Operations


class TestOperations(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.app = QApplication([])

    def setUp(self):
        self.operations = Operations()

    def test_add_operation(self):
        mock_func = Mock()
        operation = self.operations.new(
            "test_op", mock_func, "Test Operation", "path/to/icon"
        )
        self.assertIn("test_op", self.operations)
        self.assertEqual(self.operations.get("test_op"), operation)

    def test_remove_operation(self):
        mock_func = Mock()
        self.operations.new("test_op", mock_func, "Test Operation", "path/to/icon")
        self.operations.remove("test_op")
        self.assertNotIn("test_op", self.operations)

    def test_get_operation(self):
        mock_func = Mock()
        operation = self.operations.new(
            "test_op", mock_func, "Test Operation", "path/to/icon"
        )
        retrieved_operation = self.operations.get("test_op")
        self.assertEqual(operation, retrieved_operation)

    def test_new_button(self):
        mock_func = Mock()
        self.operations.new("test_op", mock_func, "Test Operation", "path/to/icon")
        button = self.operations.new_button("test_op")
        self.assertEqual(button.objectName(), "test_op")
        self.assertEqual(button.toolTip(), "Test Operation")

    def test_signal_add_operation(self):
        mock_func = Mock()
        self.operations.q_object.added.connect(mock_func)
        self.operations.new("test_op", Mock(), "Test Operation", "path/to/icon")
        mock_func.assert_called_once_with("test_op")

    def test_signal_remove_operation(self):
        mock_func = Mock()
        self.operations.new("test_op", Mock(), "Test Operation", "path/to/icon")
        self.operations.q_object.removed.connect(mock_func)
        self.operations.remove("test_op")
        mock_func.assert_called_once_with("test_op")

    @classmethod
    def tearDownClass(cls):
        cls.app.quit()


if __name__ == "__main__":
    unittest.main()



from ScopeFoundry.tools.features.new_app import new_app
from ScopeFoundry.tools.page import Page


class NewApp(Page):
    def setup(self) -> None:
        self.name = "new app"
        self.operations.new("new app", self.new_app)

        self.settings.New("cmd", str, initial="", is_cmd=True, is_clipboardable=True)

        self.operations.new("create", self.new_app)
        self.procedure = ("new app", "cmd")

    def new_app(self) -> None:
        self.settings["cmd"] = new_app()



from setuptools import setup

setup(
    name="ScopeFoundry",
    version="2.1.0",
    description="a platform for laboratory equipment control and scientific data analysis",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    # Author details
    author="Edward S. Barnard",
    author_email="esbarnard@lbl.gov",
    # Choose your license
    license="BSD",
    url="https://www.scopefoundry.org/",
    package_dir={"ScopeFoundry": "."},
    packages=[
        "ScopeFoundry",
        "ScopeFoundry.base_app",
        "ScopeFoundry.base_app.icons",
        "ScopeFoundry.controlling",
        "ScopeFoundry.controlling.pid_feedback_control_measure_docs",
        "ScopeFoundry.controlling.ranged_optimization_docs",
        "ScopeFoundry.data_browser",
        "ScopeFoundry.data_browser.viewers",
        "ScopeFoundry.data_browser.plug_ins",
        "ScopeFoundry.dynamical_widgets",
        "ScopeFoundry.examples",
        "ScopeFoundry.examples.ScopeFoundryHW",
        "ScopeFoundry.examples.ScopeFoundryHW.bsinc_noiser200",
        "ScopeFoundry.examples.ScopeFoundryHW.simulon_xyz_stage",
        "ScopeFoundry.examples.docs",
        "ScopeFoundry.examples.measurements",
        "ScopeFoundry.examples.measurements.example_2d_slowscan_measure_docs",
        "ScopeFoundry.examples.measurements.example_3d_slowscan_measure_docs",
        "ScopeFoundry.graphics",
        "ScopeFoundry.graphics.zoomable_map",
        "ScopeFoundry.logged_quantity",
        "ScopeFoundry.scanning",
        "ScopeFoundry.sequencer",
        "ScopeFoundry.sequencer.item_types",
        "ScopeFoundry.sequencer.sequencer_docs",
        "ScopeFoundry.sequencer.sweep_sequencer_docs",
        "ScopeFoundry.tools",
        "ScopeFoundry.tools.features",
        "ScopeFoundry.tools.templates",
        "ScopeFoundry.tools.pages",
    ],
    package_data={
        "": [
            "*.ui",
            "*.icns",
            "*.png",
            "*.svg",
            "*.qss",
            # for ScopeFoundry tools
            "_LICENSE",
            "*.md",
            "**/default_settings.ini",
            "**/defaults.ini",
            "**/sweep_h_centers.json",
            "**/.gitignore",
            # for docs
            "*links.json",
        ]
    },
    install_requires=["numpy", "h5py", "uuid7", "qtpy", "pyqtgraph"],
    extras_require={"all": ["PyQt6", "qtconsole", "pyqtdarktheme", "matplotlib"]},
)



import json
import unittest

from ScopeFoundry.measurement import Measurement
from ScopeFoundry import BaseMicroscopeApp, h5_io


class DummyMeasure(Measurement):

    name = "analyze_nb_test"

    def run(self):
        data = [1, 2, 4, 5]
        self.save_h5_data(data)

    def save_h5_data(self, data):
        self.h5_file = h5_io.h5_base_file(app=self.app, measurement=self)
        self.h5_meas_group = h5_io.h5_create_measurement_group(self, self.h5_file)
        self.h5_meas_group.create_dataset("count_to_4", data=data)
        self.h5_file.close()


class AnalyzeNBTest(unittest.TestCase):

    def setUp(self):
        self.app = BaseMicroscopeApp([])
        self.m = self.app.add_measurement(DummyMeasure(self.app))

    def test_non_empty_file(self):
        self.m.run()
        self.ipynb_file_name = self.app.on_analyze_with_ipynb()
        with open(self.ipynb_file_name, "r") as file:
            content = json.load(file)

        self.assertGreater(len(content["cells"]), 1)


if __name__ == "__main__":
    unittest.main()
